"""
Module which capsules the methods of the Pypeline.
"""

import collections
import configparser
import json
import multiprocessing
import os
import urllib.request
import warnings

from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.error import URLError

import h5py
import numpy as np

from typeguard import typechecked

import pynpoint

from pynpoint.core.attributes import get_attributes
from pynpoint.core.dataio import DataStorage
from pynpoint.core.processing import ProcessingModule, PypelineModule, ReadingModule, WritingModule
from pynpoint.util.module import input_info, module_info, output_info
from pynpoint.util.type_aliases import NonStaticAttribute, StaticAttribute


class Pypeline:
    """
    The :class:`~pynpoint.core.pypeline.Pypeline` class manages the pipeline modules. It inheres an
    internal dictionary of pipeline modules and has a :class:`~pynpoint.core.dataio.DataStorage`
    which is accessed by the various modules. The order in which the pipeline modules are executed
    depends on the order they have been added to the :class:`~pynpoint.core.pypeline.Pypeline`. It
    is possible to run all modules at once or run a single module by name.
    """

    @typechecked
    def __init__(self,
                 working_place_in: Optional[str] = None,
                 input_place_in: Optional[str] = None,
                 output_place_in: Optional[str] = None) -> None:
        """
        Parameters
        ----------
        working_place_in : str, None
            Working location where the central HDF5 database and the configuration file will be
            stored. Sufficient space is required in the working folder since each pipeline module
            stores a dataset in the HDF5 database. The current working folder of Python is used as
            working folder if the argument is set to None.
        input_place_in : str, None
            Default input folder where a :class:`~pynpoint.core.processing.ReadingModule` that is
            added to the :class:`~pynpoint.core.pypeline.Pypeline` will look for input data. The
            current working folder of Python is used as input folder if the argument is set to
            None.
        output_place_in : str, None
            Default output folder where a :class:`~pynpoint.core.processing.WritingModule` that is
            added to the :class:`~pynpoint.core.pypeline.Pypeline` will store output data. The
            current working folder of Python is used as output folder if the argument is set to
            None.

        Returns
        -------
        NoneType
            None
        """

        pynpoint_version = 'PynPoint v' + pynpoint.__version__

        print(len(pynpoint_version) * '=')
        print(pynpoint_version)
        print(len(pynpoint_version) * '=' + '\n')

        try:
            contents = urllib.request.urlopen('https://pypi.org/pypi/pynpoint/json').read()
            data = json.loads(contents)
            latest_version = data['info']['version']

        except URLError:
            latest_version = None

        if latest_version is not None and pynpoint.__version__ != latest_version:
            print(f'A new version ({latest_version}) is available!\n')
            print('Want to stay informed about updates, bug fixes, and new features?')
            print('Please consider using the \'Watch\' button on the Github page:')
            print('https://github.com/PynPoint/PynPoint\n')

        if working_place_in is None:
            self._m_working_place = os.getcwd()
        else:
            self._m_working_place = working_place_in

        if input_place_in is None:
            self._m_input_place = os.getcwd()
        else:
            self._m_input_place = input_place_in

        if output_place_in is None:
            self._m_output_place = os.getcwd()
        else:
            self._m_output_place = output_place_in

        print(f'Working place: {self._m_working_place}')
        print(f'Input place: {self._m_input_place}')
        print(f'Output place: {self._m_output_place}\n')

        self._m_modules = collections.OrderedDict()

        hdf5_path = os.path.join(self._m_working_place, 'PynPoint_database.hdf5')
        self.m_data_storage = DataStorage(hdf5_path)

        print(f'Database: {self.m_data_storage._m_location}')

        self._config_init()

    @typechecked
    def __setattr__(self,
                    key: str,
                    value: Any) -> None:
        """
        Internal method which assigns a value to an object attribute. This method is called
        whenever and attribute of the :class:`~pynpoint.core.pypeline.Pypeline` is changed and
        checks if the chosen working, input, or output folder exists.

        Parameters
        ----------
        key : str
            Attribute name.
        value : str
            Value for the attribute.

        Returns
        -------
        NoneType
            None
        """

        if key == '_m_working_place':
            error_msg = f'The folder that was chosen for the working place does not exist: {value}.'
            assert os.path.isdir(str(value)), error_msg

        elif key == '_m_input_place':
            error_msg = f'The folder that was chosen for the input place does not exist: {value}.'
            assert os.path.isdir(str(value)), error_msg

        elif key == '_m_output_place':
            error_msg = f'The folder that was chosen for the output place does not exist: {value}.'
            assert os.path.isdir(str(value)), error_msg

        super().__setattr__(key, value)

    @staticmethod
    @typechecked
    def _validate(module: Union[ReadingModule, WritingModule, ProcessingModule],
                  tags: List[str]) -> Tuple[bool, Optional[str]]:
        """
        Internal method to validate a :class:`~pynpoint.core.processing.PypelineModule`.

        Parameters
        ----------
        module : ReadingModule, WritingModule, ProcessingModule
            Pipeline module that will be validated.
        tags : list(str)
            Tags that are present in the database.

        Returns
        -------
        bool
            Validation of the pipeline module.
        str, None
            Pipeline module name in case it is not valid. Returns None if the module was validated.
        """

        if isinstance(module, ReadingModule):
            tags.extend(module.get_all_output_tags())

        elif isinstance(module, WritingModule):
            for tag in module.get_all_input_tags():
                if tag not in tags:
                    return False, module.name

        elif isinstance(module, ProcessingModule):
            tags.extend(module.get_all_output_tags())

            for tag in module.get_all_input_tags():
                if tag not in tags:
                    return False, module.name

        return True, None

    @typechecked
    def _config_init(self) -> None:
        """
        Internal method to initialize the configuration file. The configuration parameters are read
        from *PynPoint_config.ini* in the working folder. The file is created with default values
        (ESO/NACO) in case the file is not present.

        Returns
        -------
        NoneType
            None
        """

        @typechecked
        def _create_config(filename: str,
                           attributes: dict) -> None:

            file_obj = open(filename, 'w')
            file_obj.write('[header]\n\n')

            for key, val in attributes.items():
                if val['config'] == 'header':
                    file_obj.write(key+': '+str(val['value'])+'\n')

            file_obj.write('\n[settings]\n\n')

            for key, val in attributes.items():
                if val['config'] == 'settings':
                    file_obj.write(key+': '+str(val['value'])+'\n')

            file_obj.close()

        @typechecked
        def _read_config(config_file: str,
                         attributes: dict) -> dict:

            config = configparser.ConfigParser()

            with open(config_file) as cf_open:
                config.read_file(cf_open)

            for key, val in attributes.items():
                if config.has_option(val['config'], key):
                    if config.get(val['config'], key) == 'None':
                        if val['config'] == 'header':
                            attributes[key]['value'] = 'None'

                        # elif val['type'] == 'str':
                        #     attributes[key]['value'] = 'None'

                        elif val['type'] == 'float':
                            attributes[key]['value'] = float(0.)

                        elif val['type'] == 'int':
                            attributes[key]['value'] = int(0)

                    else:
                        if val['config'] == 'header':
                            attributes[key]['value'] = str(config.get(val['config'], key))

                        # elif val['type'] == 'str':
                        #     attributes[key]['value'] = str(config.get(val['config'], key))

                        elif val['type'] == 'float':
                            attributes[key]['value'] = float(config.get(val['config'], key))

                        elif val['type'] == 'int':
                            attributes[key]['value'] = int(config.get(val['config'], key))

            return attributes

        @typechecked
        def _write_config(attributes: dict) -> None:

            hdf = h5py.File(self._m_working_place+'/PynPoint_database.hdf5', 'a')

            if 'config' in hdf:
                del hdf['config']

            config = hdf.create_group('config')

            for key in attributes.keys():
                if attributes[key]['value'] is not None:
                    config.attrs[key] = attributes[key]['value']

            config.attrs['WORKING_PLACE'] = self._m_working_place

            hdf.close()

        config_file = os.path.join(self._m_working_place, 'PynPoint_config.ini')
        print(f'Configuration: {config_file}\n')

        attributes = get_attributes()
        attributes['CPU']['value'] = multiprocessing.cpu_count()

        if not os.path.isfile(config_file):
            warnings.warn('Configuration file not found. Creating PynPoint_config.ini with '
                          'default values in the working place.')

            _create_config(config_file, attributes)

        attributes = _read_config(config_file, attributes)

        _write_config(attributes)

        n_cpu = attributes['CPU']['value']

        if 'OMP_NUM_THREADS' in os.environ:
            n_thread = os.environ['OMP_NUM_THREADS']
        else:
            n_thread = 'not set'

        print(f'Number of CPUs: {n_cpu}')
        print(f'Number of threads: {n_thread}')

    @typechecked
    def add_module(self,
                   module: PypelineModule) -> None:
        """
        Method for adding a :class:`~pynpoint.core.processing.PypelineModule` to the internal
        dictionary of the :class:`~pynpoint.core.pypeline.Pypeline`. The module is appended at the
        end of this ordered dictionary. If the input module is a reading or writing module without
        a specified input or output location then the default location is used. The module is
        connected to the internal data storage of the :class:`~pynpoint.core.pypeline.Pypeline`.

        Parameters
        ----------
        module : ReadingModule, WritingModule, ProcessingModule
            Pipeline module that will be added to the :class:`~pynpoint.core.pypeline.Pypeline`.

        Returns
        -------
        NoneType
            None
        """

        if isinstance(module, ReadingModule):
            if module.m_input_location is None:
                module.m_input_location = self._m_input_place

        if isinstance(module, WritingModule):
            if module.m_output_location is None:
                module.m_output_location = self._m_output_place

        module.connect_database(self.m_data_storage)

        if module.name in self._m_modules:
            warnings.warn(f'Names of pipeline modules that are added to the Pypeline need to '
                          f'be unique. The current pipeline module, \'{module.name}\', does '
                          f'already exist in the Pypeline dictionary so the previous module '
                          f'with the same name will be overwritten.')

        self._m_modules[module.name] = module

    @typechecked
    def remove_module(self,
                      name: str) -> bool:
        """
        Method to remove a :class:`~pynpoint.core.processing.PypelineModule` from the internal
        dictionary with pipeline modules that are added to the
        :class:`~pynpoint.core.pypeline.Pypeline`.

        Parameters
        ----------
        name : str
            Name of the module that has to be removed.

        Returns
        -------
        bool
            Confirmation of removing the :class:`~pynpoint.core.processing.PypelineModule`.
        """

        if name in self._m_modules:
            del self._m_modules[name]

            removed = True

        else:
            warnings.warn(f'Pipeline module \'{name}\' is not found in the Pypeline dictionary '
                          f'so it could not be removed. The dictionary contains the following '
                          f'modules: {list(self._m_modules.keys())}.')

            removed = False

        return removed

    @typechecked
    def get_module_names(self) -> List[str]:
        """
        Method to return a list with the names of all pipeline modules that are added to the
        :class:`~pynpoint.core.pypeline.Pypeline`.

        Returns
        -------
        list(str)
            Ordered list of all Pypeline modules.
        """

        return list(self._m_modules.keys())

    @typechecked
    def validate_pipeline(self) -> Tuple[bool, Optional[str]]:
        """
        Method to check if each :class:`~pynpoint.core.dataio.InputPort` is pointing to an
        :class:`~pynpoint.core.dataio.OutputPort` of a previously added
        :class:`~pynpoint.core.processing.PypelineModule`.

        Returns
        -------
        bool
            Validation of the pipeline.
        str, None
            Name of the pipeline module that can not be validated. Returns None if all modules
            were validated.
        """

        self.m_data_storage.open_connection()

        # Create list with all datasets that are stored in the database
        data_tags = list(self.m_data_storage.m_data_bank.keys())

        # Initiate the validation in case self._m_modules.values() is empty
        validation = (True, None)

        # Loop over all pipline modules in the ordered dictionary
        for module in self._m_modules.values():
            # Validate the pipeline module
            validation = self._validate(module, data_tags)

            if not validation[0]:
                # Break the for loop if a module could not be validated
                break

        return validation

    @typechecked
    def validate_pipeline_module(self,
                                 name: str) -> Tuple[bool, Optional[str]]:
        """
        Method to check if each :class:`~pynpoint.core.dataio.InputPort` of a
        :class:`~pynpoint.core.processing.PypelineModule` with label ``name`` points to an
        existing dataset in the database.

        Parameters
        ----------
        name : str
            Name of the pipeline module instance that will be validated.

        Returns
        -------
        bool
            Validation of the pipeline module.
        str, None
            Pipeline module name in case it is not valid. Returns None if the module was validated.
        """

        self.m_data_storage.open_connection()

        # Create list with all datasets that are stored in the database
        data_tags = list(self.m_data_storage.m_data_bank.keys())

        # Check if the name is included in the internal dictionary with added modules
        if name in self._m_modules:
            # Validate the pipeline module
            validate = self._validate(self._m_modules[name], data_tags)

        else:
            validate = (False, name)

        return validate

    @typechecked
    def run(self) -> None:
        """
        Method for running all pipeline modules that are added to the
        :class:`~pynpoint.core.pypeline.Pypeline`.

        Returns
        -------
        NoneType
            None
        """

        # Validate the pipeline
        validation = self.validate_pipeline()

        if not validation[0]:
            # Check if the input data is available
            raise AttributeError(f'Pipeline module \'{validation[1]}\' is looking for data '
                                 f'under a tag which is not created by a previous module or '
                                 f'the data does not exist in the database.')

        # Loop over all pipeline modules and run them
        for name in self._m_modules:
            self.run_module(name)

    @typechecked
    def run_module(self,
                   name: str) -> None:
        """
        Method for running a pipeline module.

        Parameters
        ----------
        name : str
            Name of the pipeline module.

        Returns
        -------
        NoneType
            None
        """

        if name in self._m_modules:
            # Validate the pipeline module
            validation = self.validate_pipeline_module(name)

            if not validation[0]:
                raise AttributeError(f'Pipeline module \'{validation[1]}\' is looking for data '
                                     f'under a tag which does not exist in the database.')

            # Print information about the pipeline module
            module_info(self._m_modules[name])

            # Check if the module has any input ports
            if hasattr(self._m_modules[name], '_m_input_ports'):
                # Check if the list of input ports is not empty
                if len(self._m_modules[name]._m_input_ports) > 0:
                    # Print information about the input ports
                    input_info(self._m_modules[name])

            # Check if the module has any output ports
            if hasattr(self._m_modules[name], '_m_output_ports'):
                for item in self._m_modules[name]._m_output_ports:
                    # Check if the module is a ProcessingModule
                    if isinstance(self._m_modules[name], ProcessingModule):
                        # Check if the database tag is already used
                        if item in self.m_data_storage.m_data_bank:
                            # Check if the output port is not used as input port
                            if hasattr(self._m_modules[name], '_m_input_ports') and \
                                    item not in self._m_modules[name]._m_input_ports:

                                print(f'Deleting data and attributes: {item}')

                                # Delete existing data and attributes
                                self._m_modules[name]._m_output_ports[item].del_all_data()
                                self._m_modules[name]._m_output_ports[item].del_all_attributes()

            # Run the pipeline module
            self._m_modules[name].run()

            # Check if the module has any output ports
            if hasattr(self._m_modules[name], '_m_output_ports'):
                output_shape = {}

                for item in self._m_modules[name]._m_output_ports:
                    # Get the shape of the output port
                    output_shape[item] = self.get_shape(item)

                # Print information about the output ports
                output_info(self._m_modules[name], output_shape)

        else:
            warnings.warn(f'Pipeline module \'{name}\' not found.')

    @typechecked
    def get_data(self,
                 tag: str,
                 data_range: Optional[Tuple[int, int]] = None) -> np.ndarray:
        """
        Method for reading data from the database.

        Parameters
        ----------
        tag : str
            Database tag.
        data_range : tuple(int, int), None
            Slicing range for the first axis of a dataset. This argument can be used to select a
            subset of images from dataset. The full dataset is read if the argument is set to None.

        Returns
        -------
        np.ndarray
            The selected dataset from the database.
        """

        self.m_data_storage.open_connection()

        if data_range is None:
            data = np.array(self.m_data_storage.m_data_bank[tag])

        else:
            data = np.array(self.m_data_storage.m_data_bank[tag][data_range[0]:data_range[1], ])

        self.m_data_storage.close_connection()

        return data

    @typechecked
    def delete_data(self,
                    tag: str) -> None:
        """
        Method for deleting a dataset and related attributes from the central database. Disk
        space does not seem to free up when using this method.

        Parameters
        ----------
        tag : str
            Database tag.

        Returns
        -------
        NoneType
            None
        """

        self.m_data_storage.open_connection()

        if tag in self.m_data_storage.m_data_bank:
            del self.m_data_storage.m_data_bank[tag]
        else:
            warnings.warn(f'Dataset \'{tag}\' not found in the database.')

        if 'header_' + tag + '/' in self.m_data_storage.m_data_bank:
            del self.m_data_storage.m_data_bank[f'header_{tag}']
        else:
            warnings.warn(f'Attributes of \'{tag}\' not found in the database.')

        self.m_data_storage.close_connection()

    @typechecked
    def get_attribute(self,
                      data_tag: str,
                      attr_name: str,
                      static: bool = True) -> Union[StaticAttribute, NonStaticAttribute]:
        """
        Method for reading an attribute from the database.

        Parameters
        ----------
        data_tag : str
            Database tag.
        attr_name : str
            Name of the attribute.
        static : bool
            Static (True) or non-static attribute (False).

        Returns
        -------
        StaticAttribute, NonStaticAttribute
            Attribute value. For a static attribute, a single value is returned. For a non-static
            attribute, an array of values is returned.
        """

        self.m_data_storage.open_connection()

        if static:
            attr = self.m_data_storage.m_data_bank[data_tag].attrs[attr_name]

        else:
            attr = self.m_data_storage.m_data_bank[f'header_{data_tag}/{attr_name}']
            attr = np.asarray(attr)

        self.m_data_storage.close_connection()

        return attr

    @typechecked
    def set_attribute(self,
                      data_tag: str,
                      attr_name: str,
                      attr_value: Union[StaticAttribute, NonStaticAttribute],
                      static: bool = True) -> None:
        """
        Method for writing an attribute to the database. Existing values will be overwritten.

        Parameters
        ----------
        data_tag : str
            Database tag.
        attr_name : str
            Name of the attribute.
        attr_value : StaticAttribute, NonStaticAttribute
            Attribute value.
        static : bool
            Static (True) or non-static attribute (False).

        Returns
        -------
        NoneType
            None
        """

        self.m_data_storage.open_connection()

        if static:
            self.m_data_storage.m_data_bank[data_tag].attrs[attr_name] = attr_value

        else:
            if isinstance(attr_value[0], str):
                attr_value = np.array(attr_value, dtype='|S')

            if attr_name in list(self.m_data_storage.m_data_bank[f'header_{data_tag}'].keys()):
                del self.m_data_storage.m_data_bank[f'header_{data_tag}/{attr_name}']

            attr_key = f'header_{data_tag}/{attr_name}'
            self.m_data_storage.m_data_bank[attr_key] = np.asarray(attr_value)

        self.m_data_storage.close_connection()

    @typechecked
    def get_tags(self) -> List[str]:
        """
        Method for returning a list with all database tags, except header and configuration tags.

        Returns
        -------
        list(str)
            Database tags.
        """

        self.m_data_storage.open_connection()

        tags = list(self.m_data_storage.m_data_bank.keys())

        selected_tags = []

        for item in tags:
            if item in ['config', 'fits_header'] or item[0:7] == 'header_':
                continue

            selected_tags.append(item)

        self.m_data_storage.close_connection()

        return selected_tags

    @typechecked
    def get_shape(self,
                  tag: str) -> Optional[Tuple[int, ...]]:
        """
        Method for returning the shape of a database entry.

        Parameters
        ----------
        tag : str
            Database tag.

        Returns
        -------
        tuple(int, ...), None
            Shape of the dataset. None is returned if the database tag is not found.
        """

        self.m_data_storage.open_connection()

        if tag in self.m_data_storage.m_data_bank:
            data_shape = self.m_data_storage.m_data_bank[tag].shape
        else:
            data_shape = None

        self.m_data_storage.close_connection()

        return data_shape

    @typechecked
    def list_attributes(self,
                        data_tag: str) -> Dict[str, Union[str, np.float64, np.ndarray]]:
        """
        Method for printing and returning an overview of all attributes of a dataset.

        Parameters
        ----------
        data_tag : str
            Database tag of which the attributes will be extracted.

        Returns
        -------
        dict(str, bool)
            Dictionary with all attributes, both static and non-static.
        """

        print_text = f'Attribute overview of {data_tag}'

        print('\n' + len(print_text) * '-')
        print(print_text)
        print(len(print_text) * '-' + '\n')

        self.m_data_storage.open_connection()

        attributes = {}

        print('Static attributes:')

        for key, value in self.m_data_storage.m_data_bank[data_tag].attrs.items():
            attributes[key] = value
            print(f'\n   - {key} = {value}')

        print('\nNon-static attributes:')

        for key, value in self.m_data_storage.m_data_bank[f'header_{data_tag}'].items():
            attributes[key] = list(value)
            print(f'\n   - {key} = {list(value)}')

        self.m_data_storage.close_connection()

        return attributes



import os

import pytest
import numpy as np

from astropy.io import fits

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.util.tests import create_config, create_star_data, remove_test_data


class TestFitsReading:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        create_star_data(self.test_dir+'fits')
        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=['fits'], files=['filenames.dat'])

    def test_fits_reading(self) -> None:

        module = FitsReadingModule(name_in='read1',
                                   input_dir=self.test_dir+'fits',
                                   image_tag='input',
                                   overwrite=False,
                                   check=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('read1')

        data = self.pipeline.get_data('input')
        assert np.sum(data) == pytest.approx(105.54278879805277, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

    def test_fits_reading_overwrite(self) -> None:

        module = FitsReadingModule(name_in='read2',
                                   input_dir=self.test_dir+'fits',
                                   image_tag='input',
                                   overwrite=True,
                                   check=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('read2')

        data = self.pipeline.get_data('input')
        assert np.sum(data) == pytest.approx(105.54278879805277, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

    def test_static_not_found(self) -> None:

        self.pipeline.set_attribute('config', 'DIT', 'Test', static=True)

        module = FitsReadingModule(name_in='read3',
                                   input_dir=self.test_dir+'fits',
                                   image_tag='input',
                                   overwrite=True,
                                   check=True)

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('read3')

        assert len(warning) == 2

        for item in warning:
            assert item.message.args[0] == 'Static attribute DIT (=Test) not found in the FITS ' \
                                           'header.'

        self.pipeline.set_attribute('config', 'DIT', 'ESO DET DIT', static=True)

    def test_static_changing(self) -> None:

        with fits.open(self.test_dir+'fits/images_0.fits') as hdu:
            header = hdu[0].header
            header['HIERARCH ESO DET DIT'] = 0.1
            hdu.writeto(self.test_dir+'fits/images_0.fits', overwrite=True)

        with fits.open(self.test_dir+'fits/images_1.fits') as hdu:
            header = hdu[0].header
            header['HIERARCH ESO DET DIT'] = 0.2
            hdu.writeto(self.test_dir+'fits/images_1.fits', overwrite=True)

        module = FitsReadingModule(name_in='read4',
                                   input_dir=self.test_dir+'fits',
                                   image_tag='input',
                                   overwrite=True,
                                   check=True)

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('read4')

        assert len(warning) == 1

        assert warning[0].message.args[0] == f'Static attribute ESO DET DIT has changed. ' \
                                             f'Possibly the current file {self.test_dir}fits/' \
                                             f'images_1.fits does not belong to the data set ' \
                                             f'\'input\'. Attribute value is updated.'

    def test_header_attribute(self) -> None:

        with fits.open(self.test_dir+'fits/images_0.fits') as hdu:
            header = hdu[0].header
            header['PARANG'] = 1.0
            hdu.writeto(self.test_dir+'fits/images_0.fits', overwrite=True)

        with fits.open(self.test_dir+'fits/images_1.fits') as hdu:
            header = hdu[0].header
            header['PARANG'] = 2.0
            header['HIERARCH ESO DET DIT'] = 0.1
            hdu.writeto(self.test_dir+'fits/images_1.fits', overwrite=True)

        module = FitsReadingModule(name_in='read5',
                                   input_dir=self.test_dir+'fits',
                                   image_tag='input',
                                   overwrite=True,
                                   check=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('read5')

    def test_non_static_not_found(self) -> None:

        self.pipeline.set_attribute('config', 'DIT', 'None', static=True)

        for i in range(2):
            with fits.open(f'{self.test_dir}/fits/images_{i}.fits') as hdu:
                header = hdu[0].header
                del header['HIERARCH ESO DET DIT']
                del header['HIERARCH ESO DET EXP NO']
                hdu.writeto(f'{self.test_dir}/fits/images_{i}.fits', overwrite=True)

        module = FitsReadingModule(name_in='read6',
                                   input_dir=self.test_dir+'fits',
                                   image_tag='input',
                                   overwrite=True,
                                   check=True)

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('read6')

        assert len(warning) == 2

        for item in warning:
            assert item.message.args[0] == 'Non-static attribute EXP_NO (=ESO DET EXP NO) not ' \
                                           'found in the FITS header.'

    def test_fits_read_files(self) -> None:

        module = FitsReadingModule(name_in='read7',
                                   input_dir=None,
                                   image_tag='files',
                                   overwrite=False,
                                   check=True,
                                   filenames=[self.test_dir+'fits/images_0.fits',
                                              self.test_dir+'fits/images_1.fits'])

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('read7')

        assert len(warning) == 2

        for item in warning:
            assert item.message.args[0] == 'Non-static attribute EXP_NO (=ESO DET EXP NO) not ' \
                                           'found in the FITS header.'

        data = self.pipeline.get_data('files')
        assert np.sum(data) == pytest.approx(105.54278879805277, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

    def test_fits_read_textfile(self) -> None:

        with open(self.test_dir+'filenames.dat', 'w') as file_obj:
            file_obj.write(self.test_dir+'fits/images_0.fits\n')
            file_obj.write(self.test_dir+'fits/images_1.fits\n')

        module = FitsReadingModule(name_in='read8',
                                   input_dir=None,
                                   image_tag='files',
                                   overwrite=True,
                                   check=True,
                                   filenames=self.test_dir+'filenames.dat')

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('read8')

        assert len(warning) == 2

        for item in warning:
            assert item.message.args[0] == 'Non-static attribute EXP_NO (=ESO DET EXP NO) not ' \
                                           'found in the FITS header.'

        data = self.pipeline.get_data('files')
        assert np.sum(data) == pytest.approx(105.54278879805277, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

    def test_fits_read_files_exists(self) -> None:

        module = FitsReadingModule(name_in='read9',
                                   input_dir=None,
                                   image_tag='files',
                                   overwrite=True,
                                   check=True,
                                   filenames=[f'{self.test_dir}fits/images_0.fits',
                                              f'{self.test_dir}fits/images_2.fits'])

        self.pipeline.add_module(module)

        with pytest.raises(ValueError) as error:
            self.pipeline.run_module('read9')

        assert str(error.value) == f'The file {self.test_dir}fits/images_2.fits does not exist. ' \
                                   f'Please check that the path is correct.'

    def test_fits_read_textfile_exists(self) -> None:

        with open(self.test_dir+'filenames.dat', 'w') as file_obj:
            file_obj.write(self.test_dir+'fits/images_0.fits\n')
            file_obj.write(self.test_dir+'fits/images_2.fits\n')

        module = FitsReadingModule(name_in='read10',
                                   input_dir=None,
                                   image_tag='files',
                                   overwrite=True,
                                   check=True,
                                   filenames=self.test_dir+'filenames.dat')

        self.pipeline.add_module(module)

        with pytest.raises(ValueError) as error:
            self.pipeline.run_module('read10')

        assert str(error.value) == f'The file {self.test_dir}fits/images_2.fits does not exist. ' \
                                   f'Please check that the path is correct.'



import os

import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.filter import GaussianFilterModule
from pynpoint.util.tests import create_config, remove_test_data, create_star_data


class TestFilter:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        create_star_data(self.test_dir+'data')
        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=['data'])

    def test_read_data(self) -> None:

        module = FitsReadingModule(name_in='read',
                                   image_tag='data',
                                   input_dir=self.test_dir+'data',
                                   overwrite=True,
                                   check=True)

        self.pipeline.add_module(module)

        self.pipeline.run_module('read')

        data = self.pipeline.get_data('data')
        assert np.sum(data) == pytest.approx(105.54278879805277, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

    def test_gaussian_filter(self) -> None:

        module = GaussianFilterModule(name_in='filter',
                                      image_in_tag='data',
                                      image_out_tag='filtered',
                                      fwhm=0.1)

        self.pipeline.add_module(module)
        self.pipeline.run_module('filter')

        data = self.pipeline.get_data('filtered')
        assert np.sum(data) == pytest.approx(105.54278879805275, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)



"""
Interfaces for pipeline modules.
"""

import math
import os
import time
import warnings

from abc import ABCMeta, abstractmethod
from typing import Callable, List, Optional

import numpy as np

from typeguard import typechecked

from pynpoint.core.dataio import ConfigPort, DataStorage, InputPort, OutputPort
from pynpoint.util.module import progress, update_arguments
from pynpoint.util.multiline import LineProcessingCapsule
from pynpoint.util.multiproc import apply_function
from pynpoint.util.multistack import StackProcessingCapsule


class PypelineModule(metaclass=ABCMeta):
    """
    Abstract interface for the PypelineModule:

        * Reading module (:class:`pynpoint.core.processing.ReadingModule`)
        * Writing module (:class:`pynpoint.core.processing.WritingModule`)
        * Processing module (:class:`pynpoint.core.processing.ProcessingModule`)

    Each :class:`~pynpoint.core.processing.PypelineModule` has a name as a unique identifier in the
    :class:`~pynpoint.core.pypeline.Pypeline` and requires the ``connect_database`` and ``run``
    methods.
    """

    @typechecked
    def __init__(self,
                 name_in: str) -> None:
        """
        Abstract constructor of a :class:`~pynpoint.core.processing.PypelineModule`.

        Parameters
        ----------
        name_in : str
            The name of the :class:`~pynpoint.core.processing.PypelineModule`.

        Returns
        -------
        NoneType
            None
        """

        self._m_name = name_in
        self._m_data_base = None
        self._m_config_port = ConfigPort('config')

    @property
    @typechecked
    def name(self) -> str:
        """
        Returns the name of the :class:`~pynpoint.core.processing.PypelineModule`. This property
        makes sure that the internal module name can not be changed.

        Returns
        -------
        str
            The name of the :class:`~pynpoint.core.processing.PypelineModule`
        """

        return self._m_name

    @abstractmethod
    @typechecked
    def connect_database(self,
                         data_base_in: DataStorage) -> None:
        """
        Abstract interface for the function ``connect_database`` which is needed to connect a
        :class:`~pynpoint.core.dataio.Port` of a :class:`~pynpoint.core.processing.PypelineModule`
        with the :class:`~pynpoint.core.dataio.DataStorage`.

        Parameters
        ----------
        data_base_in : pynpoint.core.dataio.DataStorage
            The central database.
        """

    @abstractmethod
    @typechecked
    def run(self) -> None:
        """
        Abstract interface for the run method of :class:`~pynpoint.core.processing.PypelineModule`
        which inheres the actual algorithm behind the module.
        """


class ReadingModule(PypelineModule, metaclass=ABCMeta):
    """
    The abstract class ReadingModule is an interface for processing steps in the Pypeline which
    have only read access to the central data storage. One can specify a directory on the hard
    drive where the input data for the module is located. If no input directory is given then
    default Pypeline input directory is used. Reading modules have a dictionary of output ports
    (self._m_out_ports) but no input ports.
    """

    @typechecked
    def __init__(self,
                 name_in: str,
                 input_dir: Optional[str] = None) -> None:
        """
        Abstract constructor of ReadingModule which needs the unique name identifier as input
        (more information: :class:`pynpoint.core.processing.PypelineModule`). An input directory
        can be specified for the location of the data or else the Pypeline default directory is
        used. This function is called in all *__init__()* functions inheriting from this class.

        Parameters
        ----------
        name_in : str
            The name of the ReadingModule.
        input_dir : str
            Directory where the input files are located.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        assert (os.path.isdir(str(input_dir)) or input_dir is None), 'Input directory for ' \
            'reading module does not exist - input requested: %s.' % input_dir

        self.m_input_location = input_dir
        self._m_output_ports = {}

    @typechecked
    def add_output_port(self,
                        tag: str,
                        activation: bool = True) -> OutputPort:
        """
        Function which creates an OutputPort for a ReadingModule and appends it to the internal
        OutputPort dictionary. This function should be used by classes inheriting from
        ReadingModule to make sure that only output ports with unique tags are added. The new
        port can be used as: ::

             port = self._m_output_ports[tag]

        or by using the returned Port.

        Parameters
        ----------
        tag : str
            Tag of the new output port.
        activation : bool
            Activation status of the Port after creation. Deactivated ports will not save their
            results until they are activated.

        Returns
        -------
        pynpoint.core.dataio.OutputPort
            The new OutputPort for the ReadingModule.
        """

        port = OutputPort(tag, activate_init=activation)

        if tag in self._m_output_ports:
            warnings.warn(f'Tag \'{tag}\' of ReadingModule \'{self._m_name}\' is already used.')

        if self._m_data_base is not None:
            port.set_database_connection(self._m_data_base)

        self._m_output_ports[tag] = port

        return port

    @typechecked
    def connect_database(self,
                         data_base_in: DataStorage) -> None:
        """
        Function used by a ReadingModule to connect all ports in the internal input and output
        port dictionaries to the database. The function is called by Pypeline and connects the
        DataStorage object to all module ports.

        Parameters
        ----------
        data_base_in : pynpoint.core.dataio.DataStorage
            The central database.

        Returns
        -------
        NoneType
            None
        """

        for port in self._m_output_ports.values():
            port.set_database_connection(data_base_in)

        self._m_config_port.set_database_connection(data_base_in)

        self._m_data_base = data_base_in

    @typechecked
    def get_all_output_tags(self) -> List[str]:
        """
        Returns a list of all output tags to the ReadingModule.

        Returns
        -------
        list(str)
            List of output tags.
        """

        return list(self._m_output_ports.keys())

    @abstractmethod
    @typechecked
    def run(self) -> None:
        """
        Abstract interface for the run method of a ReadingModule which inheres the actual
        algorithm behind the module.
        """


class WritingModule(PypelineModule, metaclass=ABCMeta):
    """
    The abstract class WritingModule is an interface for processing steps in the pipeline which
    do not change the content of the internal DataStorage. They only have reading access to the
    central data base. WritingModules can be used to export data from the HDF5 database.
    WritingModules know the directory on the hard drive where the output of the module can be
    saved. If no output directory is given the default Pypeline output directory is used.
    WritingModules have a dictionary of input ports (self._m_input_ports) but no output ports.
    """

    @typechecked
    def __init__(self,
                 name_in: str,
                 output_dir: Optional[str] = None) -> None:
        """
        Abstract constructor of a WritingModule which needs the unique name identifier as input
        (more information: :class:`pynpoint.core.processing.PypelineModule`). In addition one can
        specify a output directory where the module will save its results. If no output directory is
        given the Pypeline default directory is used. This function is called in all *__init__()*
        functions inheriting from this class.

        Parameters
        ----------
        name_in : str
            The name of the WritingModule.
        output_dir : str
            Directory where the results will be saved.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        assert (os.path.isdir(str(output_dir)) or output_dir is None), 'Output directory for ' \
            'writing module does not exist - input requested: %s.' % output_dir

        self.m_output_location = output_dir
        self._m_input_ports = {}

    @typechecked
    def add_input_port(self,
                       tag: str) -> InputPort:
        """
        Function which creates an InputPort for a WritingModule and appends it to the internal
        InputPort dictionary. This function should be used by classes inheriting from WritingModule
        to make sure that only input ports with unique tags are added. The new port can be used
        as: ::

             port = self._m_input_ports[tag]

        or by using the returned Port.

        Parameters
        ----------
        tag : str
            Tag of the new input port.

        Returns
        -------
        pynpoint.core.dataio.InputPort
            The new InputPort for the WritingModule.
        """

        port = InputPort(tag)

        if self._m_data_base is not None:
            port.set_database_connection(self._m_data_base)

        self._m_input_ports[tag] = port

        return port

    @typechecked
    def connect_database(self,
                         data_base_in: DataStorage) -> None:
        """
        Function used by a WritingModule to connect all ports in the internal input and output
        port dictionaries to the database. The function is called by Pypeline and connects the
        DataStorage object to all module ports.

        Parameters
        ----------
        data_base_in : pynpoint.core.dataio.DataStorage
            The central database.

        Returns
        -------
        NoneType
            None
        """

        for port in self._m_input_ports.values():
            port.set_database_connection(data_base_in)

        self._m_config_port.set_database_connection(data_base_in)

        self._m_data_base = data_base_in

    @typechecked
    def get_all_input_tags(self) -> List[str]:
        """
        Returns a list of all input tags to the WritingModule.

        Returns
        -------
        list(str)
            List of input tags.
        """

        return list(self._m_input_ports.keys())

    @abstractmethod
    @typechecked
    def run(self) -> None:
        """
        Abstract interface for the run method of a WritingModule which inheres the actual
        algorithm behind the module.
        """


class ProcessingModule(PypelineModule, metaclass=ABCMeta):
    """
    The abstract class ProcessingModule is an interface for all processing steps in the pipeline
    which read, process, and store data. Hence processing modules have read and write access to the
    central database through a dictionary of output ports (self._m_output_ports) and a dictionary
    of input ports (self._m_input_ports).
    """

    @typechecked
    def __init__(self,
                 name_in: str) -> None:
        """
        Abstract constructor of a ProcessingModule which needs the unique name identifier as input
        (more information: :class:`pynpoint.core.processing.PypelineModule`). Call this function in
        all __init__() functions inheriting from this class.

        Parameters
        ----------
        name_in : str
             The name of the ProcessingModule.
        """

        super().__init__(name_in)

        self._m_input_ports = {}
        self._m_output_ports = {}

    @typechecked
    def add_input_port(self,
                       tag: str) -> InputPort:
        """
        Function which creates an InputPort for a ProcessingModule and appends it to the internal
        InputPort dictionary. This function should be used by classes inheriting from
        ProcessingModule to make sure that only input ports with unique tags are added. The new
        port can be used as: ::

             port = self._m_input_ports[tag]

        or by using the returned Port.

        Parameters
        ----------
        tag : str
            Tag of the new input port.

        Returns
        -------
        pynpoint.core.dataio.InputPort
            The new InputPort for the ProcessingModule.
        """

        port = InputPort(tag)

        if self._m_data_base is not None:
            port.set_database_connection(self._m_data_base)

        self._m_input_ports[tag] = port

        return port

    @typechecked
    def add_output_port(self,
                        tag: str,
                        activation: bool = True) -> OutputPort:
        """
        Function which creates an :class:`~pynpoint.core.dataio.OutputPort` for a
        :class:`~pynpoint.core.processing.ProcessingModule` and appends it to the internal
        :class:`~pynpoint.core.dataio.OutputPort` dictionary. This function should be used by
        classes inheriting from :class:`~pynpoint.core.processing.ProcessingModule` to make sure
        that only output ports with unique tags are added. The new port can be used as:

        .. code-block:: python

             port = self._m_output_ports[tag]

        or by using the returned :class:`~pynpoint.core.dataio.Port`.

        Parameters
        ----------
        tag : str
            Tag of the new output port.
        activation : bool
            Activation status of the :class:`~pynpoint.core.dataio.Port` after creation.
            Deactivated ports will not save their results until they are activated.

        Returns
        -------
        pynpoint.core.dataio.OutputPort
            The new :class:`~pynpoint.core.dataio.OutputPort` for the
            :class:`~pynpoint.core.processing.ProcessingModule`.
        """

        port = OutputPort(tag, activate_init=activation)

        if tag in self._m_output_ports:
            warnings.warn(f'Tag \'{tag}\' of ProcessingModule \'{self._m_name}\' is already used.')

        if self._m_data_base is not None:
            port.set_database_connection(self._m_data_base)

        self._m_output_ports[tag] = port

        return port

    @typechecked
    def connect_database(self,
                         data_base_in: DataStorage) -> None:
        """
        Function used by a ProcessingModule to connect all ports in the internal input and output
        port dictionaries to the database. The function is called by Pypeline and connects the
        DataStorage object to all module ports.

        Parameters
        ----------
        data_base_in : pynpoint.core.dataio.DataStorage
            The central database.

        Returns
        -------
        NoneType
            None
        """

        for port in self._m_input_ports.values():
            port.set_database_connection(data_base_in)

        for port in self._m_output_ports.values():
            port.set_database_connection(data_base_in)

        self._m_config_port.set_database_connection(data_base_in)

        self._m_data_base = data_base_in

    @typechecked
    def apply_function_in_time(self,
                               func: Callable,
                               image_in_port: InputPort,
                               image_out_port: OutputPort,
                               func_args: Optional[tuple] = None) -> None:
        """
        Applies a function to all pixel lines in time.

        Parameters
        ----------
        func : function
            The input function.
        image_in_port : pynpoint.core.dataio.InputPort
            Input port which is linked to the input data.
        image_out_port : pynpoint.core.dataio.OutputPort
            Output port which is linked to the results.
        func_args : tuple, None
            Additional arguments which are required by the input function. Not used if set to None.

        Returns
        -------
        NoneType
            None
        """

        cpu = self._m_config_port.get_attribute('CPU')

        init_line = image_in_port[:, 0, 0]

        im_shape = image_in_port.get_shape()

        size = apply_function(init_line, 0, func, func_args).shape[0]

        image_out_port.set_all(data=np.zeros((size, im_shape[1], im_shape[2])),
                               data_dim=3,
                               keep_attributes=False)

        image_in_port.close_port()
        image_out_port.close_port()

        capsule = LineProcessingCapsule(image_in_port=image_in_port,
                                        image_out_port=image_out_port,
                                        num_proc=cpu,
                                        function=func,
                                        function_args=func_args,
                                        data_length=size)

        capsule.run()

    @typechecked
    def apply_function_to_images(self,
                                 func: Callable[..., np.ndarray],
                                 image_in_port: InputPort,
                                 image_out_port: OutputPort,
                                 message: str,
                                 func_args: Optional[tuple] = None) -> None:
        """
        Function which applies a function to all images of an input port. Stacks of images are
        processed in parallel if the CPU and MEMORY attribute are set in the central configuration.
        The number of images per process is equal to the value of MEMORY divided by the value of
        CPU. Note that the function *func* is not allowed to change the shape of the images if the
        input and output port have the same tag and ``MEMORY`` is not set to None.

        Parameters
        ----------
        func : function
            The function which is applied to all images. Its definitions should be similar to::

                def function(image_in,
                             parameter1,
                             parameter2,
                             parameter3)

            The function must return a numpy array.
        image_in_port : pynpoint.core.dataio.InputPort
            Input port which is linked to the input data.
        image_out_port : pynpoint.core.dataio.OutputPort
            Output port which is linked to the results.
        message : str
            Progress message.
        func_args : tuple
            Additional arguments that are required by the input function.

        Returns
        -------
        NoneType
            None
        """

        memory = self._m_config_port.get_attribute('MEMORY')
        cpu = self._m_config_port.get_attribute('CPU')

        nimages = image_in_port.get_shape()[0]

        if memory == 0:
            memory = nimages

        if image_out_port.tag == image_in_port.tag:
            # load all images in the memory at once if the input and output tag are the
            # same or if the MEMORY attribute is set to None in the configuration file
            images = image_in_port.get_all()

            result = []

            start_time = time.time()

            for i in range(nimages):
                progress(i, nimages, message+'...', start_time)

                args = update_arguments(i, nimages, func_args)

                if args is None:
                    result.append(func(images[i, ], i))
                else:
                    result.append(func(images[i, ], i, *args))

            image_out_port.set_all(np.asarray(result), keep_attributes=True)

        elif cpu == 1:
            # process images one-by-one with a single process if CPU is set to 1
            start_time = time.time()

            for i in range(nimages):
                progress(i, nimages, message+'...', start_time)

                args = update_arguments(i, nimages, func_args)

                if args is None:
                    result = func(image_in_port[i, ], i)
                else:
                    result = func(image_in_port[i, ], i, *args)

                if result.ndim == 1:
                    image_out_port.append(result, data_dim=2)
                elif result.ndim == 2:
                    image_out_port.append(result, data_dim=3)

        else:
            # process images in parallel in stacks of MEMORY/CPU images
            print(message, end='')

            args = update_arguments(0, nimages, func_args)

            result = apply_function(image_in_port[0, :, :], 0, func, args)

            result_shape = result.shape

            out_shape = [nimages]
            for item in result_shape:
                out_shape.append(item)

            image_out_port.set_all(data=np.zeros(out_shape),
                                   data_dim=len(result_shape)+1,
                                   keep_attributes=False)

            image_in_port.close_port()
            image_out_port.close_port()

            capsule = StackProcessingCapsule(image_in_port=image_in_port,
                                             image_out_port=image_out_port,
                                             num_proc=cpu,
                                             function=func,
                                             function_args=func_args,
                                             stack_size=math.ceil(memory/cpu),
                                             result_shape=result_shape,
                                             nimages=nimages)

            capsule.run()

            print(' [DONE]')

    @typechecked
    def get_all_input_tags(self) -> List[str]:
        """
        Returns a list of all input tags to the ProcessingModule.

        Returns
        -------
        list(str)
            List of input tags.
        """

        return list(self._m_input_ports.keys())

    @typechecked
    def get_all_output_tags(self) -> List[str]:
        """
        Returns a list of all output tags to the ProcessingModule.

        Returns
        -------
        list(str)
            List of output tags.
        """

        return list(self._m_output_ports.keys())

    @abstractmethod
    @typechecked
    def run(self) -> None:
        """
        Abstract interface for the run method of a
        :class:`~pynpoint.core.processing.ProcessingModule` which inheres the actual
        algorithm behind the module.
        """



"""
Pipeline modules for frame selection.
"""

import time
import math
import warnings
import multiprocessing as mp

from typing import Optional, Tuple, Union

import numpy as np

from typeguard import typechecked
from skimage.metrics import structural_similarity, mean_squared_error

from pynpoint.core.processing import ProcessingModule
from pynpoint.util.apply_func import image_stat
from pynpoint.util.image import crop_image, pixel_distance, center_pixel
from pynpoint.util.module import progress
from pynpoint.util.remove import write_selected_data, write_selected_attributes
from pynpoint.util.star import star_positions


class RemoveFramesModule(ProcessingModule):
    """
    Pipeline module for removing images by their index number.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 selected_out_tag: str,
                 removed_out_tag: str,
                 frames: Union[str, range, list, np.ndarray]) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        selected_out_tag : str
            Tag of the database entry with the remaining images after removing the specified
            images. Should be different from *image_in_tag*.
        removed_out_tag : str
            Tag of the database entry with the images that are removed. Should be different
            from *image_in_tag*.
        frames : str, list, range, numpy.ndarray
            The frame indices that have to be removed or a database tag pointing to a list of
            frame indices.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)

        self.m_selected_out_port = self.add_output_port(selected_out_tag)
        self.m_removed_out_port = self.add_output_port(removed_out_tag)

        if isinstance(frames, str):
            self.m_index_in_port = self.add_input_port(frames)

        else:
            self.m_index_in_port = None

            if isinstance(frames, (range, list)):
                self.m_frames = np.asarray(frames, dtype=int)

            elif isinstance(frames, np.ndarray):
                self.m_frames = frames

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Removes the frames and corresponding attributes, updates the
        NFRAMES attribute, and saves the data and attributes.

        Returns
        -------
        NoneType
            None
        """

        if self.m_index_in_port is not None:
            self.m_frames = self.m_index_in_port.get_all()

        if np.size(np.where(self.m_frames >= self.m_image_in_port.get_shape()[0])) > 0:
            raise ValueError(f'Some values in \'frames\' are larger than the total number of '
                             f'available frames, {self.m_image_in_port.get_shape()[0]}.')

        write_selected_data(memory=self._m_config_port.get_attribute('MEMORY'),
                            indices=self.m_frames,
                            image_in_port=self.m_image_in_port,
                            selected_out_port=self.m_selected_out_port,
                            removed_out_port=self.m_removed_out_port)

        write_selected_attributes(indices=self.m_frames,
                                  image_in_port=self.m_image_in_port,
                                  selected_out_port=self.m_selected_out_port,
                                  removed_out_port=self.m_removed_out_port,
                                  module_type='RemoveFramesModule',
                                  history=f'frames removed = {np.size(self.m_frames)}')

        self.m_image_in_port.close_port()


class FrameSelectionModule(ProcessingModule):
    """
    Pipeline module for applying a frame selection.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 selected_out_tag: str,
                 removed_out_tag: str,
                 index_out_tag: Optional[str] = None,
                 method: str = 'median',
                 threshold: Union[float, Tuple[float, float]] = 4.,
                 fwhm: Optional[float] = 0.1,
                 aperture: Union[Tuple[str, float], Tuple[str, float, float]] = ('circular', 0.2),
                 position: Optional[Union[Tuple[int, int, float], Tuple[None, None, float],
                                    Tuple[int, int, None]]] = None) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        selected_out_tag : str
            Tag of the database entry with the selected images that are
            written as output. Should be different from
            ``image_in_tag``.
        removed_out_tag : str
            Tag of the database entry with the removed images that are
            written as output. Should be different from
            ``image_in_tag``.
        index_out_tag : str, None
            Tag of the database entry with the list of frames indices
            that are removed with the frames selection. No data is
            written when set to ``None``.
        method : str
            Method used for the frame selection. Either sigma clipping
            is applied with respect to the median (``method='median'``)
            or maximum (``method='max'``) aperture flux or fluxes
            outside a specified range (``method='range'``) are removed.
            In the latter case, the ``threshold`` argument should be
            a tuple with the minimum and maximum flux value.
        threshold : float, tuple(float, float)
            Threshold in units of sigma for the frame selection in case
            ``method='median'`` or ``method='max'``. All images that are
            a ``threshold`` number of sigmas away from the
            median/maximum aperture flux will be removed. In case
            ``method='range'``, the argument should be a tuple with the
            minimum and maximum flux. Aperture fluxes within this range
            will be selected and stored at ``selected_out_tag``.
        fwhm : float, None
            The FWHM (arcsec) of the Gaussian kernel that is used to
            smooth the images before the brightest pixel is located.
            Recommended to be similar in size to the FWHM of the
            stellar PSF. No smoothing is applied if set to ``None``.
        aperture : tuple(str, float), tuple(str, float, float)
            Tuple with the aperture properties for measuring the
            photometry around the location of the brightest pixel. The
            first element contains the aperture type ('circular',
            'annulus', or 'ratio'). For a circular aperture, the second
            element contains the aperture radius (arcsec). For the
            other two types, the second and third element are the
            inner and outer radii (arcsec) of the aperture.
        position : tuple(int, int, float), None
            Subframe that is selected to search for the star. The tuple
            contains the center (pix) and size (arcsec) (pos_x, pos_y,
            size). Setting ``position`` to None will use the full image
            to search for the star. If `position=(None, None, size)`
            then the center of the image will be used. If
            ``position=(pos_x, pos_y, None)`` then a fixed position is
            used for the aperture.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)

        self.m_selected_out_port = self.add_output_port(selected_out_tag)
        self.m_removed_out_port = self.add_output_port(removed_out_tag)

        if index_out_tag is None:
            self.m_index_out_port = None
        else:
            self.m_index_out_port = self.add_output_port(index_out_tag)

        self.m_method = method
        self.m_fwhm = fwhm
        self.m_aperture = aperture
        self.m_threshold = threshold
        self.m_position = position

    @staticmethod
    @typechecked
    def aperture_phot(image: np.ndarray,
                      position: np.ndarray,
                      aperture: Union[Tuple[str, float, float],
                                      Tuple[str, None, float]]) -> np.float64:
        """
        Parameters
        ----------
        image : np.ndarray
            Input image (2D).
        position : np.ndarray
            Center position (y, x) of the aperture.
        aperture : tuple(str, float, float)
            Tuple with the aperture properties for measuring the
            photometry around the location of the brightest pixel. The
            first element contains the aperture type ('circular',
            'annulus', or 'ratio'). For a circular aperture, the second
            element is empty and the third element contains the aperture
            radius (pix). For the other two types, the second and third
            element are the inner and outer radii (pix) of the aperture.

        Returns
        -------
        np.float64
            Photometry value.
        """

        check_pos_in = any(np.floor(position[:]-aperture[2]) < 0.)
        check_pos_out = any(np.ceil(position[:]+aperture[2]) > image.shape[0])

        if check_pos_in or check_pos_out:
            phot = np.nan

        else:
            im_crop = crop_image(image, tuple(position), 2*int(math.ceil(aperture[2])))

            npix = im_crop.shape[0]

            x_grid = y_grid = np.linspace(-(npix-1)/2, (npix-1)/2, npix)
            xx_grid, yy_grid = np.meshgrid(x_grid, y_grid)
            rr_grid = np.sqrt(xx_grid**2 + yy_grid**2)

            if aperture[0] == 'circular':
                phot = np.sum(im_crop[rr_grid < aperture[2]])

            elif aperture[0] == 'annulus':
                phot = np.sum(im_crop[(rr_grid > aperture[1]) & (rr_grid < aperture[2])])

            elif aperture[0] == 'ratio':
                phot = np.sum(im_crop[rr_grid < aperture[1]]) / \
                       np.sum(im_crop[(rr_grid > aperture[1]) & (rr_grid < aperture[2])])

        return phot

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Smooths the images with a Gaussian
        kernel, locates the brightest pixel in each image, measures the
        integrated flux around the brightest pixel, calculates the
        median and standard deviation of the photometry, and applies
        sigma clipping to remove low quality images.

        Returns
        -------
        NoneType
            None
        """

        pixscale = self.m_image_in_port.get_attribute('PIXSCALE')
        nimages = self.m_image_in_port.get_shape()[0]

        if self.m_fwhm is not None:
            self.m_fwhm = int(math.ceil(self.m_fwhm/pixscale))

        if self.m_position is not None and self.m_position[2] is not None:
            self.m_position = (self.m_position[0],
                               self.m_position[0],
                               int(math.ceil(self.m_position[2]/pixscale)))

        if len(self.m_aperture) == 2:
            self.m_aperture = (self.m_aperture[0],
                               None,
                               self.m_aperture[1]/pixscale)

        elif len(self.m_aperture) == 3:
            self.m_aperture = (self.m_aperture[0],
                               self.m_aperture[1]/pixscale,
                               self.m_aperture[2]/pixscale)

        starpos = star_positions(self.m_image_in_port, self.m_fwhm, self.m_position)

        phot = np.zeros(nimages)
        start_time = time.time()

        for i in range(nimages):
            progress(i, nimages, 'Aperture photometry...', start_time)

            phot[i] = self.aperture_phot(self.m_image_in_port[i, ], starpos[i, :], self.m_aperture)

        if self.m_method == 'median':
            phot_ref = np.nanmedian(phot)
            print(f'Median = {phot_ref:.2f}')

        elif self.m_method == 'max':
            phot_ref = np.nanmax(phot)
            print(f'Maximum = {phot_ref:.2f}')

        elif self.m_method == 'range':
            phot_ref = np.nanmedian(phot)
            print(f'Median = {phot_ref:.2f}')

        phot_std = np.nanstd(phot)
        print(f'Standard deviation = {phot_std:.2f}')

        if self.m_method in ['median', 'max']:
            index_del = np.logical_or((phot > phot_ref + self.m_threshold*phot_std),
                                      (phot < phot_ref - self.m_threshold*phot_std))

        elif self.m_method == 'range':
            index_del = np.logical_or((phot < self.m_threshold[0]),
                                      (phot > self.m_threshold[1]))

        index_del[np.isnan(phot)] = True
        index_del = np.where(index_del)[0]

        index_sel = np.ones(nimages, dtype=bool)
        index_sel[index_del] = False

        write_selected_data(memory=self._m_config_port.get_attribute('MEMORY'),
                            indices=index_del,
                            image_in_port=self.m_image_in_port,
                            selected_out_port=self.m_selected_out_port,
                            removed_out_port=self.m_removed_out_port)

        history = f'frames removed = {np.size(index_del)}'

        if self.m_index_out_port is not None:
            self.m_index_out_port.set_all(index_del, data_dim=1)
            self.m_index_out_port.copy_attributes(self.m_image_in_port)
            self.m_index_out_port.add_attribute('STAR_POSITION', starpos, static=False)
            self.m_index_out_port.add_history('FrameSelectionModule', history)

        write_selected_attributes(indices=index_del,
                                  image_in_port=self.m_image_in_port,
                                  selected_out_port=self.m_selected_out_port,
                                  removed_out_port=self.m_removed_out_port,
                                  module_type='FrameSelectionModule',
                                  history=history)

        self.m_selected_out_port.add_attribute('STAR_POSITION', starpos[index_sel], static=False)
        self.m_selected_out_port.add_history('FrameSelectionModule', history)

        self.m_removed_out_port.add_attribute('STAR_POSITION', starpos[index_del], static=False)
        self.m_removed_out_port.add_history('FrameSelectionModule', history)

        self.m_image_in_port.close_port()


class RemoveLastFrameModule(ProcessingModule):
    """
    Pipeline module for removing every NDIT+1 image from NACO dataset obtained in cube mode. This
    frame contains the average pixel values of the cube.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry that is written as output.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Removes every NDIT+1 frame and saves the data and attributes.

        Returns
        -------
        NoneType
            None
        """

        ndit = self.m_image_in_port.get_attribute('NDIT')
        nframes = self.m_image_in_port.get_attribute('NFRAMES')
        index = self.m_image_in_port.get_attribute('INDEX')

        nframes_new = []
        index_new = []

        start_time = time.time()
        for i, item in enumerate(ndit):
            progress(i, len(ndit), 'Removing the last image of each FITS cube...', start_time)

            if nframes[i] != item+1:
                warnings.warn(f'Number of frames ({nframes[i]}) is not equal to NDIT+1.')

            frame_start = np.sum(nframes[0:i])
            frame_end = np.sum(nframes[0:i+1]) - 1

            nframes_new.append(nframes[i]-1)
            index_new.extend(index[frame_start:frame_end])

            self.m_image_out_port.append(self.m_image_in_port[frame_start:frame_end, ])

        nframes_new = np.asarray(nframes_new, dtype=int)
        index_new = np.asarray(index_new, dtype=int)

        self.m_image_out_port.copy_attributes(self.m_image_in_port)

        self.m_image_out_port.add_attribute('NFRAMES', nframes_new, static=False)
        self.m_image_out_port.add_attribute('INDEX', index_new, static=False)

        history = 'frames removed = NDIT+1'
        self.m_image_out_port.add_history('RemoveLastFrameModule', history)

        self.m_image_out_port.close_port()


class RemoveStartFramesModule(ProcessingModule):
    """
    Pipeline module for removing a fixed number of images at the beginning of each cube. This can
    be useful for NACO data in which the background is higher at the beginning of the cube.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 frames: int = 1) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry that is written as output.
        frames : int
            Number of frames that are removed at the beginning of each cube.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_frames = int(frames)

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Removes a constant number of images at the beginning of each cube
        and saves the data and attributes.

        Returns
        -------
        NoneType
            None
        """

        if self.m_image_out_port.tag == self.m_image_in_port.tag:
            raise ValueError('Input and output port should have a different tag.')

        nframes = self.m_image_in_port.get_attribute('NFRAMES')
        index = self.m_image_in_port.get_attribute('INDEX')

        index_new = []

        if 'PARANG' in self.m_image_in_port.get_all_non_static_attributes():
            parang = self.m_image_in_port.get_attribute('PARANG')
            parang_new = []

        else:
            parang = None

        if 'STAR_POSITION' in self.m_image_in_port.get_all_non_static_attributes():
            star = self.m_image_in_port.get_attribute('STAR_POSITION')
            star_new = []

        else:
            star = None

        start_time = time.time()
        for i, _ in enumerate(nframes):
            progress(i, len(nframes), 'Removing images at the begin of each cube...', start_time)

            frame_start = np.sum(nframes[0:i]) + self.m_frames
            frame_end = np.sum(nframes[0:i+1])

            if frame_start >= frame_end:
                raise ValueError('The number of frames in the original data cube is equal or '
                                 'smaller than the number of frames that have to be removed.')

            index_new.extend(index[frame_start:frame_end])

            if parang is not None:
                parang_new.extend(parang[frame_start:frame_end])

            if star is not None:
                star_new.extend(star[frame_start:frame_end])

            self.m_image_out_port.append(self.m_image_in_port[frame_start:frame_end, ])

        self.m_image_out_port.copy_attributes(self.m_image_in_port)

        self.m_image_out_port.add_attribute('NFRAMES', nframes-self.m_frames, static=False)
        self.m_image_out_port.add_attribute('INDEX', index_new, static=False)

        if parang is not None:
            self.m_image_out_port.add_attribute('PARANG', parang_new, static=False)

        if star is not None:
            self.m_image_out_port.add_attribute('STAR_POSITION', np.asarray(star_new), static=False)

        history = 'frames removed = '+str(self.m_frames)
        self.m_image_out_port.add_history('RemoveStartFramesModule', history)

        self.m_image_out_port.close_port()


class ImageStatisticsModule(ProcessingModule):
    """
    Pipeline module for calculating image statistics for the full images or a subsection of the
    images.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 stat_out_tag: str,
                 position: Optional[Union[Tuple[int, int, float],
                                          Tuple[None, None, float]]] = None) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry with the images that are read as input.
        stat_out_tag : str
            Tag of the database entry with the statistical results that are written as output. The
            result is stored in the following order: minimum, maximum, sum, mean, median, and
            standard deviation.
        position : tuple(int, int, float)
            Position (x, y) (pix) and radius (arcsec) of the circular area in which the statistics
            are calculated. The full image is used if set to None.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_stat_out_port = self.add_output_port(stat_out_tag)

        self.m_position = position

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Calculates the minimum, maximum, sum, mean, median, and standard
        deviation of the pixel values of each image separately. NaNs are ignored for each
        calculation. The values are calculated for either the full images or a circular
        subsection of the images.

        Returns
        -------
        NoneType
            None
        """

        pixscale = self.m_image_in_port.get_attribute('PIXSCALE')

        nimages = self.m_image_in_port.get_shape()[0]
        im_shape = self.m_image_in_port.get_shape()[1:]

        if self.m_position is None:
            indices = None

        else:
            if self.m_position[0] is None and self.m_position[1] is None:
                center = center_pixel(self.m_image_in_port[0, ])

                self.m_position = (center[0],  # y position
                                   center[1],  # x position
                                   self.m_position[2]/pixscale)  # radius (pix)

            else:
                self.m_position = (int(self.m_position[1]),  # y position
                                   int(self.m_position[0]),  # x position
                                   self.m_position[2]/pixscale)  # radius (pix)

            rr_grid, _, _ = pixel_distance(im_shape, position=self.m_position[0:2])
            rr_reshape = np.reshape(rr_grid, (rr_grid.shape[0]*rr_grid.shape[1]))
            indices = np.where(rr_reshape <= self.m_position[2])[0]

        self.apply_function_to_images(image_stat,
                                      self.m_image_in_port,
                                      self.m_stat_out_port,
                                      'Calculating image statistics',
                                      func_args=(indices, ))

        history = f'number of images = {nimages}'
        self.m_stat_out_port.copy_attributes(self.m_image_in_port)
        self.m_stat_out_port.add_history('ImageStatisticsModule', history)
        self.m_stat_out_port.close_port()


class FrameSimilarityModule(ProcessingModule):
    """
    Pipeline module for measuring the similarity between frames.
    """

    __author__ = 'Benedikt Schmidhuber, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_tag: str,
                 method: str = 'MSE',
                 mask_radius: Tuple[float, float] = (0., 5.),
                 window_size: float = 0.1,
                 temporal_median: str = 'full') -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_tag : str
            Tag of the database entry that is read as input.
        method : str
            Method for the similarity measure. There are three measures available:

                - `MSE` - Mean Squared Error
                - `PCC` - Pearson Correlation Coefficient
                - `SSIM` - Structural Similarity

            These measures compare each image to the temporal median of the image set.
        mask_radius : tuple(float, float)
            Inner and outer radius (arcsec) of the mask that is applied to the images.
            The inner radius is actually not used so can be set to zero.
        window_size : float
            Size (arcsec) of the sliding window that is used when the SSIM similarity is
            calculated.
        temporal_median : str
            Option to calculate the temporal median for each position ('full') or as a constant
            value ('constant') for the entire set. The latter is computationally less expensive.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_tag)
        self.m_image_out_port = self.add_output_port(image_tag)

        if method not in ('MSE', 'PCC', 'SSIM'):
            raise ValueError(f'The chosen method \'{method}\' is not available. Please ensure '
                             f'that you have selected one of \'MSE\', \'PCC\', or \'SSIM\'.')

        if temporal_median not in ('full', 'constant'):
            raise ValueError(f'The chosen temporal_median \'{temporal_median}\' is not '
                             f'available. Please ensure that you have selected one of \'full\', '
                             f'\'constant\'.')

        self.m_method = method
        self.m_temporal_median = temporal_median
        self.m_mask_radii = mask_radius
        self.m_window_size = window_size

    @staticmethod
    @typechecked
    def _similarity(images: np.ndarray,
                    reference_index: int,
                    mode: str,
                    window_size: int,
                    temporal_median: Union[bool, np.ndarray] = False) -> Tuple[int, float]:
        """
        Internal function to compute the MSE as defined by Ruane et al. 2019.
        """

        @typechecked
        def _temporal_median(reference_index: int,
                             images: np.ndarray) -> np.ndarray:
            """
            Internal function to calculate the temporal median for all frames, except the one with
            the ``reference_index``.
            """

            image_m = np.concatenate((images[:reference_index], images[reference_index+1:]))

            return np.median(image_m, axis=0)

        image_x_i = images[reference_index]

        if isinstance(temporal_median, bool):
            image_m = _temporal_median(reference_index, images=images)
        else:
            image_m = temporal_median

        if mode == 'MSE':
            return reference_index, mean_squared_error(image_x_i, image_m)

        if mode == 'PCC':
            # calculate the covariance matrix of the flattened images
            cov_mat = np.cov(image_x_i.flatten(), image_m.flatten(), ddof=1)

            # the variances are stored in the diagonal, therefore take the sqrt to obtain std
            std = np.sqrt(np.diag(cov_mat))

            # does not matter whether [0, 1] or [1, 0] as cov_mat is symmetric
            return reference_index, cov_mat[0, 1] / (std[0] * std[1])

        if mode == 'SSIM':
            # winsize needs to be odd
            if int(window_size) % 2 == 0:
                winsize = int(window_size) + 1
            else:
                winsize = int(window_size)

            # TODO Unclear what value to pass to data_range
            # Previously the argument was not requires
            data_range = np.amax(image_x_i)-np.amin(image_x_i)

            struc_sim = structural_similarity(im1=image_x_i,
                                              im2=image_m,
                                              win_size=winsize,
                                              data_range=data_range)

            return reference_index, struc_sim

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Computes the similarity between frames based on the Mean Squared
        Error (MSE), the Pearson Correlation Coefficient (PCC), or the Structural Similarity (SSIM).
        The correlation values are stored as non-static attribute (``MSE``, ``PCC``, or ``SSIM``)
        to the input data. After running this module, the
        :class:`~pynpoint.processing.frameselection.SelectByAttributeModule` can be used to select
        the images with the highest correlation.

        Returns
        -------
        NoneType
            None
        """

        cpu = self._m_config_port.get_attribute('CPU')
        pixscale = self.m_image_in_port.get_attribute('PIXSCALE')

        # get number of images
        nimages = self.m_image_in_port.get_shape()[0]

        # convert arcsecs to pixels
        self.m_mask_radii = (math.floor(self.m_mask_radii[0] / pixscale),
                             math.floor(self.m_mask_radii[1] / pixscale))
        self.m_window_size = int(self.m_window_size / pixscale)

        # overlay the same mask over all images
        images = self.m_image_in_port.get_all()

        # close the port during the calculations
        self.m_image_out_port.close_port()

        # Change the radius to the image size
        images = crop_image(images, None, int(2.*self.m_mask_radii[1]))

        if self.m_temporal_median == 'constant':
            temporal_median = np.median(images, axis=0)
        else:
            temporal_median = False

        # compare images and store similarity
        similarities = np.zeros(nimages)

        pool = mp.Pool(cpu)
        async_results = []

        for i in range(nimages):
            async_results.append(pool.apply_async(FrameSimilarityModule._similarity,
                                                  args=(images,
                                                        i,
                                                        self.m_method,
                                                        self.m_window_size,
                                                        temporal_median)))

        pool.close()

        start_time = time.time()

        # wait for all processes to finish
        while mp.active_children():
            # number of finished processes
            nfinished = sum([i.ready() for i in async_results])

            progress(nfinished, nimages, 'Calculating image similarity...', start_time)

            # check if new processes have finished every 5 seconds
            time.sleep(5)

        if nfinished != nimages:
            print('\r                                                      ')
            print('\rCalculating image similarity... [DONE]')

        # get the results for every async_result object
        for async_result in async_results:
            reference, similarity = async_result.get()
            similarities[reference] = similarity

        pool.terminate()

        # reopen the port after the calculation
        self.m_image_out_port.open_port()
        self.m_image_out_port.add_attribute(f'{self.m_method}', similarities, static=False)
        self.m_image_out_port.close_port()


class SelectByAttributeModule(ProcessingModule):
    """
    Pipeline module for selecting frames based on attribute values.
    """

    __author__ = 'Benedikt Schmidhuber, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 selected_out_tag: str,
                 removed_out_tag: str,
                 attribute_tag: str,
                 number_frames: int = 100,
                 order: str = 'descending') -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        selected_out_tag : str
            Tag of the database entry to which the selected frames are written.
        removed_out_tag : str
            Tag of the database entry to which the removed frames are written.
        attribute_tag : str
            Name of the attribute which is used to sort and select the frames.
        number_frames : int
            Number of frames that are selected.
        order : str
            Order in which the frames are selected. Can be either 'descending' (will select the
            lowest attribute values) or 'ascending' (will select the highest attribute values).

        Returns
        -------
        NoneType
            None

        Examples
        --------
        The example below selects the first 100 frames with an ascending order of the ``INDEX``
        values that are stored to the 'im_arr' dataset::

            SelectByAttributeModule(name_in='frame_selection',
                                    image_in_tag='im_arr',
                                    attribute_tag='INDEX',
                                    selected_frames=100,
                                    order='ascending',
                                    selected_out_tag='im_arr_selected',
                                    removed_out_tag='im_arr_removed'))

        The example below selects the 200 frames with the largest ``SSIM`` values that
        are stored to the 'im_arr' dataset::

            SelectByAttributeModule(name_in='frame_selection',
                                    image_in_tag='im_arr',
                                    attribute_tag='SSIM',
                                    selected_frames=200,
                                    order='descending',
                                    selected_out_tag='im_arr_selected',
                                    removed_out_tag='im_arr_removed'))
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_selected_out_port = self.add_output_port(selected_out_tag)
        self.m_removed_out_port = self.add_output_port(removed_out_tag)

        if order not in ('ascending', 'descending'):
            raise ValueError('The selected order is not available. The available options are '
                             '\'ascending\' or \'descending\'.')

        self.m_attribute_tag = attribute_tag
        self.m_number_frames = number_frames
        self.m_order = order

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Selects images according to a specified attribute tag and
        ordering, e.g. the highest 150 ``INDEX`` frames, or the lowest 50 ``PCC`` frames.
        The order of the selected images is determined by the `descending` or `ascending`
        attribute values. To sort the images again by their original order, the
        :class:`~pynpoint.processing.psfpreparation.SortParangModule` can be used.

        Returns
        -------
        NoneType
            None
        """

        nimages = self.m_image_in_port.get_shape()[0]
        attribute = self.m_image_in_port.get_attribute(f'{self.m_attribute_tag}')

        if nimages != attribute.size:
            raise ValueError(f'The attribute {{self.m_attribute_tag}} does not have the same '
                             f'length ({len(attribute)}) as the tag has images ({nimages}). '
                             f'Please check the attribute you have chosen for selection.')

        if self.m_order == 'descending':
            # sort attribute in descending order
            sorting_order = np.argsort(attribute)[::-1]
        else:
            # sort attribute in ascending order
            sorting_order = np.argsort(attribute)

        index_del = sorting_order[self.m_number_frames:]

        write_selected_data(memory=self._m_config_port.get_attribute('MEMORY'),
                            indices=index_del,
                            image_in_port=self.m_image_in_port,
                            selected_out_port=self.m_selected_out_port,
                            removed_out_port=self.m_removed_out_port)

        write_selected_attributes(indices=index_del,
                                  image_in_port=self.m_image_in_port,
                                  selected_out_port=self.m_selected_out_port,
                                  removed_out_port=self.m_removed_out_port,
                                  module_type='SelectByAttributeModule',
                                  history=f'selected tag = {self.m_attribute_tag}')

        self.m_image_in_port.close_port()


class ResidualSelectionModule(ProcessingModule):
    """
    Pipeline module for applying a frame selection on the residuals of the PSF subtraction.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 selected_out_tag: str,
                 removed_out_tag: str,
                 percentage: float = 10.,
                 annulus_radii: Tuple[float, float] = (0.1, 0.2)) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        selected_out_tag : str
            Tag of the database entry with the selected images that are written as output.
        removed_out_tag : str
            Tag of the database entry with the removed images that are written as output.
        percentage : float
            The percentage of best frames that is selected.
        annulus_radii : tuple(float, float)
            Inner and outer radius of the annulus (arcsec).

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)

        self.m_selected_out_port = self.add_output_port(selected_out_tag)
        self.m_removed_out_port = self.add_output_port(removed_out_tag)

        self.m_percentage = percentage
        self.m_annulus_radii = annulus_radii

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Applies a frame selection on the derotated residuals from the
        PSF subtraction. The pixels within an annulus (e.g. at the separation of an expected
        planet) are selected and the standard deviation is calculated. The chosen percentage
        of images with the lowest standard deviation are stored as output.

        Returns
        -------
        NoneType
            None
        """

        pixscale = self.m_image_in_port.get_attribute('PIXSCALE')
        nimages = self.m_image_in_port.get_shape()[0]
        npix = self.m_image_in_port.get_shape()[-1]

        rr_grid, _, _ = pixel_distance((npix, npix), position=None)

        pixel_select = np.where((rr_grid > self.m_annulus_radii[0]/pixscale) &
                                (rr_grid < self.m_annulus_radii[1]/pixscale))

        start_time = time.time()
        phot_annulus = np.zeros(nimages)

        for i in range(nimages):
            progress(i, nimages, 'Aperture photometry...', start_time)

            phot_annulus[i] = np.sum(np.abs(self.m_image_in_port[i][pixel_select]))

        print(f'Minimum, maximum = {np.amin(phot_annulus):.2f}, {np.amax(phot_annulus):.2f}')
        print(f'Mean, median = {np.nanmean(phot_annulus):.2f}, {np.nanmedian(phot_annulus):.2f}')
        print(f'Standard deviation = {np.nanstd(phot_annulus):.2f}')

        n_select = int(nimages*self.m_percentage/100.)
        index_del = np.argsort(phot_annulus)[n_select:]

        write_selected_data(memory=self._m_config_port.get_attribute('MEMORY'),
                            indices=index_del,
                            image_in_port=self.m_image_in_port,
                            selected_out_port=self.m_selected_out_port,
                            removed_out_port=self.m_removed_out_port)

        write_selected_attributes(indices=index_del,
                                  image_in_port=self.m_image_in_port,
                                  selected_out_port=self.m_selected_out_port,
                                  removed_out_port=self.m_removed_out_port,
                                  module_type='ResidualSelectionModule',
                                  history=f'frames removed = {index_del.size}')

        self.m_image_in_port.close_port()



import warnings

from pynpoint.core.pypeline import Pypeline

from pynpoint.processing.background import SimpleBackgroundSubtractionModule, \
                                           MeanBackgroundSubtractionModule, \
                                           LineSubtractionModule, \
                                           NoddingBackgroundModule

from pynpoint.processing.badpixel import BadPixelSigmaFilterModule, \
                                         BadPixelInterpolationModule, \
                                         BadPixelMapModule, \
                                         BadPixelTimeFilterModule, \
                                         ReplaceBadPixelsModule

from pynpoint.processing.basic import SubtractImagesModule, \
                                      AddImagesModule, \
                                      RotateImagesModule, \
                                      RepeatImagesModule

from pynpoint.processing.centering import StarAlignmentModule, \
                                          FitCenterModule, \
                                          ShiftImagesModule, \
                                          WaffleCenteringModule

from pynpoint.processing.darkflat import DarkCalibrationModule, \
                                         FlatCalibrationModule

from pynpoint.processing.extract import StarExtractionModule, \
                                        ExtractBinaryModule

from pynpoint.processing.filter import GaussianFilterModule

from pynpoint.processing.fluxposition import FakePlanetModule, \
                                             SimplexMinimizationModule, \
                                             FalsePositiveModule, \
                                             MCMCsamplingModule, \
                                             AperturePhotometryModule, \
                                             SystematicErrorModule

from pynpoint.processing.frameselection import RemoveFramesModule, \
                                               FrameSelectionModule, \
                                               RemoveLastFrameModule, \
                                               RemoveStartFramesModule, \
                                               ImageStatisticsModule, \
                                               FrameSimilarityModule, \
                                               SelectByAttributeModule, \
                                               ResidualSelectionModule

from pynpoint.processing.limits import ContrastCurveModule, \
                                       MassLimitsModule

from pynpoint.processing.pcabackground import PCABackgroundPreparationModule, \
                                              PCABackgroundSubtractionModule, \
                                              DitheringBackgroundModule

from pynpoint.processing.psfpreparation import PSFpreparationModule, \
                                               AngleInterpolationModule, \
                                               AngleCalculationModule, \
                                               SortParangModule, \
                                               SDIpreparationModule

from pynpoint.processing.psfsubtraction import PcaPsfSubtractionModule, \
                                               ClassicalADIModule

from pynpoint.processing.resizing import CropImagesModule, \
                                         ScaleImagesModule, \
                                         AddLinesModule, \
                                         RemoveLinesModule

from pynpoint.processing.stacksubset import StackAndSubsetModule, \
                                            StackCubesModule, \
                                            DerotateAndStackModule, \
                                            CombineTagsModule

from pynpoint.processing.timedenoising import CwtWaveletConfiguration, \
                                              DwtWaveletConfiguration, \
                                              WaveletTimeDenoisingModule, \
                                              TimeNormalizationModule

from pynpoint.readwrite.attr_reading import AttributeReadingModule, \
                                            ParangReadingModule, \
                                            WavelengthReadingModule

from pynpoint.readwrite.attr_writing import AttributeWritingModule, \
                                            ParangWritingModule

from pynpoint.readwrite.fitsreading import FitsReadingModule

from pynpoint.readwrite.fitswriting import FitsWritingModule

from pynpoint.readwrite.hdf5reading import Hdf5ReadingModule

from pynpoint.readwrite.hdf5writing import Hdf5WritingModule

from pynpoint.readwrite.textwriting import TextWritingModule

from pynpoint.readwrite.nearreading import NearReadingModule

warnings.simplefilter('always', DeprecationWarning)

__author__ = 'Tomas Stolker & Markus Bonse'
__license__ = 'GPLv3'
__version__ = '0.11.0'
__maintainer__ = 'Tomas Stolker'
__email__ = 'stolker@strw.leidenuniv.nl'
__status__ = 'Development'



import os

import h5py
import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.hdf5reading import Hdf5ReadingModule
from pynpoint.util.tests import create_config, create_random, remove_test_data


class TestHdf5Reading:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        create_random(self.test_dir+'data')
        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=['data'])

    def test_hdf5_reading(self) -> None:

        data = np.random.normal(loc=0, scale=2e-4, size=(4, 10, 10))

        with h5py.File(self.test_dir+'data/PynPoint_database.hdf5', 'a') as hdf_file:
            hdf_file.create_dataset('extra', data=data)
            hdf_file.create_dataset('header_extra/PARANG', data=[1., 2., 3., 4.])

        module = Hdf5ReadingModule(name_in='read1',
                                   input_filename='PynPoint_database.hdf5',
                                   input_dir=self.test_dir+'data',
                                   tag_dictionary={'images': 'images'})

        self.pipeline.add_module(module)
        self.pipeline.run_module('read1')

        data = self.pipeline.get_data('images')
        assert np.sum(data) == pytest.approx(0.007153603490533874, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

    def test_dictionary_none(self) -> None:

        module = Hdf5ReadingModule(name_in='read2',
                                   input_filename='PynPoint_database.hdf5',
                                   input_dir=self.test_dir+'data',
                                   tag_dictionary=None)

        self.pipeline.add_module(module)
        self.pipeline.run_module('read2')

        data = self.pipeline.get_data('images')
        assert np.sum(data) == pytest.approx(0.007153603490533874, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

    def test_wrong_tag(self) -> None:

        module = Hdf5ReadingModule(name_in='read3',
                                   input_filename='PynPoint_database.hdf5',
                                   input_dir=self.test_dir+'data',
                                   tag_dictionary={'test': 'test'})

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('read3')

        assert len(warning) == 1
        assert warning[0].message.args[0] == 'The dataset with tag name \'test\' is not found in ' \
                                             'the HDF5 file.'

        with h5py.File(self.test_dir+'data/PynPoint_database.hdf5', 'r') as hdf_file:
            assert set(hdf_file.keys()) == set(['extra', 'header_extra', 'header_images', 'images'])

    def test_no_input_filename(self) -> None:

        module = Hdf5ReadingModule(name_in='read4',
                                   input_filename=None,
                                   input_dir=self.test_dir+'data',
                                   tag_dictionary=None)

        self.pipeline.add_module(module)
        self.pipeline.run_module('read4')

        data = self.pipeline.get_data('images')
        assert np.sum(data) == pytest.approx(0.007153603490533874, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)



"""
Functions for testing the pipeline and its modules.
"""

import os
import math
import shutil
import subprocess

from typing import List, Optional, Tuple, Union

import h5py
import numpy as np

from typeguard import typechecked
from astropy.io import fits
from scipy.ndimage import shift


@typechecked
def create_config(filename: str) -> None:
    """
    Create a configuration file.

    Parameters
    ----------
    filename : str
        Configuration filename.

    Returns
    -------
    NoneType
        None
    """

    with open(filename, 'w') as file_obj:

        file_obj.write('[header]\n\n')
        file_obj.write('INSTRUMENT: INSTRUME\n')
        file_obj.write('NFRAMES: NAXIS3\n')
        file_obj.write('EXP_NO: ESO DET EXP NO\n')
        file_obj.write('NDIT: ESO DET NDIT\n')
        file_obj.write('PARANG_START: ESO ADA POSANG\n')
        file_obj.write('PARANG_END: ESO ADA POSANG END\n')
        file_obj.write('DITHER_X: ESO SEQ CUMOFFSETX\n')
        file_obj.write('DITHER_Y: ESO SEQ CUMOFFSETY\n')
        file_obj.write('DIT: None\n')
        file_obj.write('LATITUDE: None\n')
        file_obj.write('LONGITUDE: None\n')
        file_obj.write('PUPIL: None\n')
        file_obj.write('DATE: None\n')
        file_obj.write('RA: None\n')
        file_obj.write('DEC: None\n\n')
        file_obj.write('[settings]\n\n')
        file_obj.write('PIXSCALE: 0.027\n')
        file_obj.write('MEMORY: 39\n')
        file_obj.write('CPU: 1\n')


@typechecked
def create_random(path: str,
                  nimages: float = 5) -> None:
    """
    Create a dataset of images with Gaussian distributed pixel values.

    Parameters
    ----------
    path : str
        Working folder.
    nimages : int
        Number of images.

    Returns
    -------
    NoneType
        None
    """

    if not os.path.exists(path):
        os.makedirs(path)

    file_in = os.path.join(path, 'PynPoint_database.hdf5')

    np.random.seed(1)
    images = np.random.normal(loc=0, scale=2e-4, size=(nimages, 11, 11))

    with h5py.File(file_in, 'w') as h5_file:
        dset = h5_file.create_dataset('images', data=images)
        dset.attrs['PIXSCALE'] = 0.01

        h5_file.create_dataset('header_images/PARANG', data=np.arange(float(nimages)))


@typechecked
def create_fits(path: str,
                filename: str,
                image: np.ndarray,
                ndit: int,
                exp_no: int,
                dither_x: float,
                dither_y: float) -> None:
    """
    Create a FITS file with images and header information.

    Parameters
    ----------
    path : str
        Working folder.
    filename : str
        FITS filename.
    image : np.ndarray
        Images.
    ndit : int
        Number of integrations.
    exp_no : int
        Exposure number.
    dither_x : float
        Horizontal dither position relative to the image center.
    dither_y : float
        Vertical dither position relative to the image center.

    Returns
    -------
    NoneType
        None
    """

    hdu = fits.PrimaryHDU()
    header = hdu.header
    header['INSTRUME'] = 'IMAGER'
    header['HIERARCH ESO DET EXP NO'] = 1.
    header['HIERARCH ESO DET NDIT'] = ndit
    header['HIERARCH ESO DET EXP NO'] = exp_no
    header['HIERARCH ESO ADA POSANG'] = 0.
    header['HIERARCH ESO ADA POSANG END'] = 180.
    header['HIERARCH ESO SEQ CUMOFFSETX'] = dither_x
    header['HIERARCH ESO SEQ CUMOFFSETY'] = dither_y
    hdu.data = image
    hdu.writeto(os.path.join(path, filename))


@typechecked
def create_fake_data(path: str) -> None:
    """
    Create an ADI dataset with a star and planet.

    Parameters
    ----------
    path : str
        Working folder.

    Returns
    -------
    NoneType
        None
    """

    if not os.path.exists(path):
        os.makedirs(path)

    ndit = 10
    npix = 21
    fwhm = 3.
    sep = 6.
    contrast = 1e-1
    pos_star = 10.
    exp_no = 1
    parang = np.linspace(0., 180., 10)

    np.random.seed(1)

    sigma = fwhm / (2.*math.sqrt(2.*math.log(2.)))

    x = np.arange(0., 21., 1.)
    y = np.arange(0., 21., 1.)

    xx, yy = np.meshgrid(x, y)

    images = np.zeros((ndit, npix, npix))

    for i, item in enumerate(parang):
        images[i, ] = np.random.normal(loc=0, scale=2e-4, size=(npix, npix))

        star = np.exp(-((xx-pos_star)**2+(yy-pos_star)**2)/(2.*sigma**2))/(2.*np.pi*sigma**2)

        x_shift = sep*math.cos(math.radians(item))
        y_shift = sep*math.sin(math.radians(item))

        images[i, ] += star + shift(contrast*star, (x_shift, y_shift), order=5)

    create_fits(path, 'images.fits', images, ndit, exp_no, 0., 0.)


@typechecked
def create_ifs_data(path: str) -> None:
    """
    Create an IFS dataset with a star and planet.

    Parameters
    ----------
    path : str
        Working folder.

    Returns
    -------
    NoneType
        None
    """

    ndit = 10
    npix = 21
    nwavel = 3
    fwhm = 3.
    sep = 6.
    contrast = 1.
    pos_star = 10.
    exp_no = 1

    parang = np.linspace(0., 180., 10)
    wavelength = [1., 1.1, 1.2]

    if not os.path.exists(path):
        os.makedirs(path)

    sigma = fwhm / (2.*math.sqrt(2.*math.log(2.)))

    x = y = np.arange(0., 21., 1.)
    xx, yy = np.meshgrid(x, y)

    np.random.seed(1)

    images = np.random.normal(loc=0, scale=0.05, size=(nwavel, ndit, npix, npix))

    for i, par_item in enumerate(parang):
        for j, wav_item in enumerate(wavelength):
            sigma_scale = sigma*wav_item

            star = np.exp(-((xx-pos_star)**2+(yy-pos_star)**2)/(2.*sigma_scale**2))

            x_shift = sep*math.cos(math.radians(par_item))
            y_shift = sep*math.sin(math.radians(par_item))

            images[j, i, ] += star + shift(contrast*star, (x_shift, y_shift), order=5)

    create_fits(path, 'images.fits', images, ndit, exp_no, 0., 0.)


@typechecked
def create_star_data(path: str,
                     npix: int = 11,
                     pos_star: float = 5.) -> None:
    """
    Create a dataset with a PSF and Gaussian noise.

    Parameters
    ----------
    path : str
        Working folder.
    npix : int
        Number of pixels in each dimension.

    Returns
    -------
    NoneType
        None
    """

    fwhm = 3.
    nimages = 5

    exp_no = [1, 2]
    parang_start = [0., 90.]
    parang_end = [90., 180.]

    if not os.path.exists(path):
        os.makedirs(path)

    np.random.seed(1)

    for j, item in enumerate(exp_no):
        sigma = fwhm / (2. * math.sqrt(2.*math.log(2.)))

        x = y = np.arange(0., float(npix), 1.)
        xx, yy = np.meshgrid(x, y)

        images = np.random.normal(loc=0, scale=0.1, size=(nimages, npix, npix))
        images += np.exp(-((xx-pos_star)**2+(yy-pos_star)**2)/(2.*sigma**2))

        hdu = fits.PrimaryHDU()
        header = hdu.header
        header['INSTRUME'] = 'IMAGER'
        header['HIERARCH ESO DET EXP NO'] = item
        header['HIERARCH ESO DET NDIT'] = nimages
        header['HIERARCH ESO ADA POSANG'] = parang_start[j]
        header['HIERARCH ESO ADA POSANG END'] = parang_end[j]
        header['HIERARCH ESO SEQ CUMOFFSETX'] = 'None'
        header['HIERARCH ESO SEQ CUMOFFSETY'] = 'None'
        hdu.data = images
        hdu.writeto(os.path.join(path, f'images_{j}.fits'))


@typechecked
def create_dither_data(path: str) -> None:
    """
    Create a dithering dataset with a stellar PSF.

    Parameters
    ----------
    path : str
        Working folder.

    Returns
    -------
    NoneType
        None
    """

    if not os.path.exists(path):
        os.makedirs(path)

    ndit = 5
    npix = 21
    fwhm = 3.

    exp_no = [1, 2, 3, 4]
    pos_star = [(5., 5.), (5., 15.), (15., 15.), (15., 5.)]
    parang = np.full(10, 0.)

    np.random.seed(1)

    sigma = fwhm / (2.*math.sqrt(2.*math.log(2.)))

    x = np.arange(0., 21., 1.)
    y = np.arange(0., 21., 1.)

    xx, yy = np.meshgrid(x, y)

    for i, item in enumerate(exp_no):
        images = np.random.normal(loc=0, scale=0.1, size=(ndit, npix, npix))

        for j in range(ndit):
            images[j, ] += np.exp(-((xx-pos_star[i][0])**2+(yy-pos_star[i][1])**2)/(2.*sigma**2))

        create_fits(path, f'images_{i}.fits', images, ndit, item,
                    pos_star[i][0]-10., pos_star[i][1]-10.)


@typechecked
def create_waffle_data(path: str) -> None:
    """
    Create data with satellite spots and Gaussian noise.

    Parameters
    ----------
    path : str
        Working folder.

    Returns
    -------
    NoneType
        None
    """

    if not os.path.exists(path):
        os.makedirs(path)

    fwhm = 3
    npix = 101

    x_spot = [25., 25., 75., 75.]
    y_spot = [25., 75., 75., 25.]

    sigma = fwhm / (2. * math.sqrt(2.*math.log(2.)))

    x = y = np.arange(0., npix, 1.)
    xx, yy = np.meshgrid(x, y)

    image = np.zeros((npix, npix))

    for j in range(4):
        image += np.exp(-((xx-x_spot[j])**2+(yy-y_spot[j])**2)/(2.*sigma**2))/(2.*np.pi*sigma**2)

    hdu = fits.PrimaryHDU()
    header = hdu.header
    header['INSTRUME'] = 'IMAGER'
    header['HIERARCH ESO DET EXP NO'] = 'None'
    header['HIERARCH ESO DET NDIT'] = 'None'
    header['HIERARCH ESO ADA POSANG'] = 'None'
    header['HIERARCH ESO ADA POSANG END'] = 'None'
    header['HIERARCH ESO SEQ CUMOFFSETX'] = 'None'
    header['HIERARCH ESO SEQ CUMOFFSETY'] = 'None'
    hdu.data = image
    hdu.writeto(os.path.join(path, 'images.fits'))


@typechecked
def remove_test_data(path: str,
                     folders: Optional[List[str]] = None,
                     files: Optional[List[str]] = None) -> None:
    """
    Function to remove data created by the test cases.

    Parameters
    ----------
    path : str
        Working folder.
    folders : list(str, )
        Folders to remove.
    files : list(str, )
        Files to removes.

    Returns
    -------
    NoneType
        None
    """

    os.remove(path+'PynPoint_database.hdf5')
    os.remove(path+'PynPoint_config.ini')

    if folders is not None:
        for item in folders:
            shutil.rmtree(path+item)

    if files is not None:
        for item in files:
            os.remove(path+item)


@typechecked
def create_near_data(path: str) -> None:
    """
    Create a stack of images with Gaussian distributed pixel values.

    Parameters
    ----------
    path : str
        Working folder.

    Returns
    -------
    NoneType
        None
    """

    if not os.path.exists(path):
        os.makedirs(path)

    np.random.seed(1)

    image = np.random.normal(loc=0., scale=1., size=(10, 10))

    exp_no = [1, 2]

    for i, item in enumerate(exp_no):
        fits_file = os.path.join(path, f'images_{i}.fits')

        primary_header = fits.Header()
        primary_header['INSTRUME'] = 'VISIR'
        primary_header['HIERARCH ESO DET CHOP NCYCLES'] = 5
        primary_header['HIERARCH ESO DET SEQ1 DIT'] = 1.
        primary_header['HIERARCH ESO TPL EXPNO'] = item
        primary_header['HIERARCH ESO DET CHOP ST'] = 'T'
        primary_header['HIERARCH ESO DET CHOP CYCSKIP'] = 0
        primary_header['HIERARCH ESO DET CHOP CYCSUM'] = 'F'

        chopa_header = fits.Header()
        chopa_header['HIERARCH ESO DET FRAM TYPE'] = 'HCYCLE1'

        chopb_header = fits.Header()
        chopb_header['HIERARCH ESO DET FRAM TYPE'] = 'HCYCLE2'

        hdu = [fits.PrimaryHDU(header=primary_header)]

        for _ in range(5):
            hdu.append(fits.ImageHDU(image, header=chopa_header))
            hdu.append(fits.ImageHDU(image, header=chopb_header))

        # last image is the average of all images
        hdu.append(fits.ImageHDU(image))

        hdulist = fits.HDUList(hdu)
        hdulist.writeto(fits_file)

        subprocess.call('compress '+fits_file, shell=True)



"""
Functions for calculating detection limits.
"""

import math

from typing import Tuple

import numpy as np
from scipy.stats import t

from photutils.aperture import aperture_photometry, CircularAperture
from typeguard import typechecked

from pynpoint.util.analysis import student_t, fake_planet,\
    compute_aperture_flux_elements
from pynpoint.util.image import polar_to_cartesian, center_subpixel
from pynpoint.util.psf import pca_psf_subtraction
from pynpoint.util.residuals import combine_residuals


@typechecked
def contrast_limit(path_images: str,
                   path_psf: str,
                   noise: np.ndarray,
                   mask: np.ndarray,
                   parang: np.ndarray,
                   psf_scaling: float,
                   extra_rot: float,
                   pca_number: int,
                   threshold: Tuple[str, float],
                   aperture: float,
                   residuals: str,
                   snr_inject: float,
                   position: Tuple[float, float]) -> Tuple[float, float, float, float]:

    """
    Function for calculating the contrast limit at a specified position for a given sigma level or
    false positive fraction, both corrected for small sample statistics.

    Parameters
    ----------
    path_images : str
        System location of the stack of images (3D).
    path_psf : str
        System location of the PSF template for the fake planet (3D). Either a single image or a
        stack of images equal in size to science data.
    noise : numpy.ndarray
        Residuals of the PSF subtraction (3D) without injection of fake planets. Used to measure
        the noise level with a correction for small sample statistics.
    mask : numpy.ndarray
        Mask (2D).
    parang : numpy.ndarray
        Derotation angles (deg).
    psf_scaling : float
        Additional scaling factor of the planet flux (e.g., to correct for a neutral density
        filter). Should have a positive value.
    extra_rot : float
        Additional rotation angle of the images in clockwise direction (deg).
    pca_number : int
        Number of principal components used for the PSF subtraction.
    threshold : tuple(str, float)
        Detection threshold for the contrast curve, either in terms of 'sigma' or the false
        positive fraction (FPF). The value is a tuple, for example provided as ('sigma', 5.) or
        ('fpf', 1e-6). Note that when sigma is fixed, the false positive fraction will change with
        separation. Also, sigma only corresponds to the standard deviation of a normal distribution
        at large separations (i.e., large number of samples).
    aperture : float
        Aperture radius (pix) for the calculation of the false positive fraction.
    residuals : str
        Method used for combining the residuals ('mean', 'median', 'weighted', or 'clipped').
    snr_inject : float
        Signal-to-noise ratio of the injected planet signal that is used to measure the amount
        of self-subtraction.
    position : tuple(float, float)
        The separation (pix) and position angle (deg) of the fake planet.

    Returns
    -------
    float
        Separation (pix).
    float
        Position angle (deg).
    float
        Contrast (mag).
    float
        False positive fraction.
    """

    images = np.load(path_images)
    psf = np.load(path_psf)

    # Cartesian coordinates of the fake planet
    yx_fake = polar_to_cartesian(images, position[0], position[1]-extra_rot)

    # Determine the noise level
    noise_apertures = compute_aperture_flux_elements(image=noise[0, ],
                                                     x_pos=yx_fake[1],
                                                     y_pos=yx_fake[0],
                                                     size=aperture,
                                                     ignore=False)

    t_noise = np.std(noise_apertures, ddof=1) * \
        math.sqrt(1 + 1 / (noise_apertures.shape[0]))

    # get sigma from fpf or fpf from sigma
    # Note that the number of degrees of freedom is given by nu = n-1 with n the number of samples.
    # See Section 3 of Mawet et al. (2014) for more details on the Student's t distribution.

    if threshold[0] == 'sigma':
        sigma = threshold[1]

        # Calculate the FPF for a given sigma level

        fpf = t.sf(sigma, noise_apertures.shape[0] - 1,
                   loc=0., scale=1.)

    elif threshold[0] == 'fpf':
        fpf = threshold[1]

        # Calculate the sigma level for a given FPF
        sigma = t.isf(fpf, noise_apertures.shape[0] - 1,
                      loc=0., scale=1.)

    else:
        raise ValueError('Threshold type not recognized.')

    # Aperture properties
    im_center = center_subpixel(images)

    # Measure the flux of the star
    ap_phot = CircularAperture((im_center[1], im_center[0]), aperture)
    phot_table = aperture_photometry(psf_scaling*psf[0, ], ap_phot, method='exact')
    star = phot_table['aperture_sum'][0]

    # Magnitude of the injected planet
    flux_in = snr_inject*t_noise
    mag = -2.5*math.log10(flux_in/star)

    # Inject the fake planet
    fake = fake_planet(images=images,
                       psf=psf,
                       parang=parang,
                       position=(position[0], position[1]),
                       magnitude=mag,
                       psf_scaling=psf_scaling)

    # Run the PSF subtraction
    _, im_res = pca_psf_subtraction(images=fake*mask,
                                    angles=-1.*parang+extra_rot,
                                    pca_number=pca_number)

    # Stack the residuals
    im_res = combine_residuals(method=residuals, res_rot=im_res)
    flux_out_frame = im_res[0, ] - noise[0, ]

    # Measure the flux of the fake planet after PCA
    # the first element is the planet
    flux_out = compute_aperture_flux_elements(image=flux_out_frame,
                                              x_pos=yx_fake[1],
                                              y_pos=yx_fake[0],
                                              size=aperture,
                                              ignore=False)[0]

    # Calculate the amount of self-subtraction
    attenuation = flux_out/flux_in
    # the throughput can not be negative. However, this can happen due to numerical inaccuracies
    if attenuation < 0:
        attenuation = 0

    # Calculate the detection limit
    contrast = (sigma*t_noise + np.mean(noise_apertures))/(attenuation*star)

    # The flux_out can be negative, for example if the aperture includes self-subtraction regions
    if contrast > 0.:
        contrast = -2.5*math.log10(contrast)
    else:
        contrast = np.nan

    # Separation [pix], position angle [deg], contrast [mag], FPF
    return position[0], position[1], contrast, fpf



"""
Abstract interfaces for multiprocessing applications with the poison pill pattern.
"""

import multiprocessing
from sys import platform

from typing import Callable, Optional, Union
from abc import ABCMeta, abstractmethod

import numpy as np

from typeguard import typechecked

from pynpoint.core.dataio import InputPort, OutputPort


# On macOS, the spawn start method is the default since Python 3.8.
# The fork start method should be considered unsafe as it can lead
# to crashes of the subprocess according to the documentation.
if platform in ['darwin', 'linux']:
    multiprocessing.set_start_method('fork')


class TaskInput:
    """
    Class for tasks that are processed by the :class:`~pynpoint.util.multiproc.TaskProcessor`.
    """

    @typechecked
    def __init__(self,
                 input_data: Union[np.ndarray, np.integer, tuple],
                 job_parameter: tuple) -> None:
        """
        Parameters
        ----------
        input_data : int, float, np.ndarray
            Input data for by the :class:`~pynpoint.util.multiproc.TaskProcessor`.
        job_parameter : tuple
            Additional data or parameters.

        Returns
        -------
        NoneType
            None
        """

        self.m_input_data = input_data
        self.m_job_parameter = job_parameter


class TaskResult:
    """
    Class for results that can be stored by the :class:`~pynpoint.util.multiproc.TaskWriter`.
    """

    @typechecked
    def __init__(self,
                 data_array: np.ndarray,
                 position: tuple) -> None:
        """
        Parameters
        ----------
        data_array : np.ndarray
            Array with the results for a given position.
        position : tuple(tuple(int, int, int), tuple(int, int, int), tuple(int, int, int))
             The position where the results will be stored.

        Returns
        -------
        NoneType
            None
        """

        self.m_data_array = data_array
        self.m_position = position


class TaskCreator(multiprocessing.Process, metaclass=ABCMeta):
    """
    Abstract interface for :class:`~pynpoint.util.multiproc.TaskCreator` classes. A
    :class:`~pynpoint.util.multiproc.TaskCreator` creates instances of
    :class:`~pynpoint.util.multiproc.TaskInput`, which will be processed by the
    :class:`~pynpoint.util.multiproc.TaskProcessor`, and appends them to the central task
    queue. In general there is only one :class:`~pynpoint.util.multiproc.TaskCreator` running
    for a poison pill multiprocessing application. A :class:`~pynpoint.util.multiproc.TaskCreator`
    communicates with to the :class:`~pynpoint.util.multiproc.TaskWriter` in order to avoid
    simultaneously access to the central database.
    """

    @typechecked
    def __init__(self,
                 data_port_in: Optional[InputPort],
                 tasks_queue_in: multiprocessing.JoinableQueue,
                 data_mutex_in: Optional[multiprocessing.Lock],
                 num_proc: int) -> None:
        """
        Parameters
        ----------
        data_port_in : pynpoint.core.dataio.InputPort, None
            An input port which links to the data that has to be processed.
        tasks_queue_in : multiprocessing.queues.JoinableQueue
            The central task queue.
        data_mutex_in : multiprocessing.synchronize.Lock, None
            A mutex shared with the writer to ensure that no read and write operations happen at
            the same time.
        num_proc : int
            Maximum number of instances of :class:`~pynpoint.util.multiproc.TaskProcessor` that run
            simultaneously.

        Returns
        -------
        NoneType
            None
        """

        multiprocessing.Process.__init__(self)

        self.m_data_in_port = data_port_in
        self.m_task_queue = tasks_queue_in
        self.m_data_mutex = data_mutex_in
        self.m_num_proc = num_proc

    @typechecked
    def create_poison_pills(self) -> None:
        """
        Creates poison pills for the :class:`~pynpoint.util.multiproc.TaskProcessor` and
        :class:`~pynpoint.util.multiproc.TaskWriter`. A process will shut down if it receives a
        poison pill as a new task. This method should be executed at the end of the
        :func:`~pynpoint.util.multiproc.TaskCreator.run` method.

        Returns
        -------
        NoneType
            None
        """

        for _ in range(self.m_num_proc-1):
            # poison pills
            self.m_task_queue.put(1)

        # final poison pill
        self.m_task_queue.put(None)

    @abstractmethod
    @typechecked
    def run(self) -> None:
        """
        Creates objects of the :class:`~pynpoint.util.multiproc.TaskInput` until all tasks are
        placed in the task queue.

        Returns
        -------
        NoneType
            None
        """


class TaskProcessor(multiprocessing.Process, metaclass=ABCMeta):
    """
    Abstract interface for :class:`~pynpoint.util.multiproc.TaskProcessor` classes. The number of
    instances of :class:`~pynpoint.util.multiproc.TaskProcessor` that run simultaneously in a
    poison pill multiprocessing application can be set with ``CPU`` parameter in the central
    configuration file. A :class:`~pynpoint.util.multiproc.TaskProcessor` takes tasks from a task
    queue, processes the task, and stores the results back into a result queue. The process will
    shut down if the next task is a poison pill. The order in which process finish is not fixed.
    """

    @typechecked
    def __init__(self,
                 tasks_queue_in: multiprocessing.JoinableQueue,
                 result_queue_in: multiprocessing.JoinableQueue) -> None:
        """
        Parameters
        ----------
        tasks_queue_in : multiprocessing.queues.JoinableQueue
            The input task queue with instances of :class:`~pynpoint.util.multiproc.TaskInput`.
        result_queue_in : multiprocessing.queues.JoinableQueue
            The result task queue with instances of :class:`~pynpoint.util.multiproc.TaskResult`.

        Returns
        -------
        NoneType
            None
        """

        multiprocessing.Process.__init__(self)

        self.m_task_queue = tasks_queue_in
        self.m_result_queue = result_queue_in

    @typechecked
    def check_poison_pill(self,
                          next_task: Union[TaskInput, int, None]) -> bool:
        """
        Function to check if the next task is a poison pill.

        Parameters
        ----------
        next_task : int, None, pynpoint.util.multiproc.TaskInput
            The next task.

        Returns
        -------
        bool
            True if the next task is a poison pill, False otherwise.
        """

        if next_task == 1:
            # poison pill
            poison_pill = True

            self.m_task_queue.task_done()

        elif next_task is None:
            # final poison pill
            poison_pill = True

            # shut down writer process
            self.m_result_queue.put(None)
            self.m_task_queue.task_done()

        else:
            # no poison pill
            poison_pill = False

        return poison_pill

    @typechecked
    def run(self) -> None:
        """
        Run method to start the :class:`~pynpoint.util.multiproc.TaskProcessor`. The run method
        will continue to process tasks from the input task queue until it receives a poison pill.

        Returns
        -------
        NoneType
            None
        """

        while True:
            next_task = self.m_task_queue.get()

            if self.check_poison_pill(next_task):
                break

            result = self.run_job(next_task)

            self.m_task_queue.task_done()
            self.m_result_queue.put(result)

    @abstractmethod
    @typechecked
    def run_job(self,
                tmp_task: TaskInput) -> None:
        """
        Abstract interface for the :func:`~pynpoint.util.multiproc.TaskProcessor.run_job` method
        which is called from the :func:`~pynpoint.util.multiproc.TaskProcessor.run` method for each
        task individually.

        Parameters
        ----------
        tmp_task : pynpoint.util.multiproc.TaskInput
            Input task.

        Returns
        -------
        NoneType
            None
        """


class TaskWriter(multiprocessing.Process):
    """
    The :class:`~pynpoint.util.multiproc.TaskWriter` receives results from the result queue, which
    have been computed by a :class:`~pynpoint.util.multiproc.TaskProcessor`, and stores the results
    in the central database. The position parameter of the
    :class:`~pynpoint.util.multiproc.TaskResult` is used to slice the result to the correct
    position in the complete output dataset.
    """

    @typechecked
    def __init__(self,
                 result_queue_in: multiprocessing.JoinableQueue,
                 data_out_port_in: Optional[OutputPort],
                 data_mutex_in: multiprocessing.Lock) -> None:
        """
        Parameters
        ----------
        result_queue_in : multiprocessing.queues.JoinableQueue
            The result queue.
        data_out_port_in : pynpoint.core.dataio.OutputPort, None
            The output port where the results will be stored.
        data_mutex_in : multiprocessing.synchronize.Lock
            A mutex that is shared with the :class:`~pynpoint.util.multiproc.TaskWriter` which
            ensures that read and write operations to the database do not occur simultaneously.

        Returns
        -------
        NoneType
            None
        """

        multiprocessing.Process.__init__(self)

        self.m_result_queue = result_queue_in
        self.m_data_mutex = data_mutex_in
        self.m_data_out_port = data_out_port_in

    @typechecked
    def check_poison_pill(self,
                          next_result: Union[TaskResult, None]) -> int:
        """
        Function to check if the next result is a poison pill.

        Parameters
        ----------
        next_result : None, pynpoint.util.multiproc.TaskResult
            The next result.

        Returns
        -------
        int
            0 -> no poison pill, 1 -> poison pill, 2 -> poison pill but still results in the
            queue (rare error case).
        """

        if next_result is None:
            # check if there are results after the poison pill
            if self.m_result_queue.empty():
                poison_pill = 1

                # shut down the writer
                self.m_result_queue.task_done()

            else:
                poison_pill = 2

                # put back the poison pill for the moment
                self.m_result_queue.put(None)
                self.m_result_queue.task_done()

        else:
            poison_pill = 0

        return poison_pill

    @typechecked
    def run(self) -> None:
        """
        Run method of the :class:`~pynpoint.util.multiproc.TaskWriter`. It is called once when
        it has to start storing the results until it receives a poison pill.

        Returns
        -------
        NoneType
            None
        """

        while True:
            next_result = self.m_result_queue.get()
            poison_pill_case = self.check_poison_pill(next_result)

            if poison_pill_case == 1:
                break

            if poison_pill_case == 2:
                continue

            with self.m_data_mutex:
                self.m_data_out_port._check_status_and_activate()
                self.m_data_out_port[to_slice(next_result.m_position)] = next_result.m_data_array
                self.m_data_out_port.close_port()

            self.m_result_queue.task_done()


class MultiprocessingCapsule(metaclass=ABCMeta):
    """
    Abstract interface for multiprocessing capsules based on the poison pill pattern.
    """

    @typechecked
    def __init__(self,
                 image_in_port: Optional[InputPort],
                 image_out_port: Optional[OutputPort],
                 num_proc: int) -> None:
        """
        Parameters
        ----------
        image_in_port : pynpoint.core.dataio.InputPort, None
            Port to the input data.
        image_out_port : pynpoint.core.dataio.OutputPort, None
            Port to the place where the output data will be stored.
        num_proc : int
            Number of task processors.

        Returns
        -------
        NoneType
            None
        """

        # buffer twice the data as processes are available
        self.m_tasks_queue = multiprocessing.JoinableQueue(maxsize=num_proc)
        self.m_result_queue = multiprocessing.JoinableQueue(maxsize=num_proc)
        self.m_num_proc = num_proc

        # database mutex
        self.m_data_mutex = multiprocessing.Lock()

        # create reader
        self.m_creator = self.init_creator(image_in_port)

        # create processors
        self.m_task_processors = self.create_processors()

        # create writer
        self.m_writer = self.create_writer(image_out_port)

    @abstractmethod
    @typechecked
    def create_processors(self) -> None:
        """
        Function that is called from the constructor to create a list of instances of
        :class:`~pynpoint.util.multiproc.TaskProcessor`.

        Returns
        -------
        NoneType
            None
        """

    @abstractmethod
    @typechecked
    def init_creator(self,
                     image_in_port: Optional[InputPort]) -> None:
        """
        Function that is called from the constructor to create a
        :class:`~pynpoint.util.multiproc.TaskCreator`.

        Parameters
        ----------
        image_in_port : pynpoint.core.dataio.InputPort, None
            Input port for the task creator.

        Returns
        -------
        NoneType
            None
        """

    @typechecked
    def create_writer(self,
                      image_out_port: Optional[OutputPort]) -> TaskWriter:
        """
        Function that is called from the constructor to create the
        :class:`~pynpoint.util.multiproc.TaskWriter`.

        Parameters
        ----------
        image_out_port : pynpoint.core.dataio.OutputPort, None
            Output port for the creator.

        Returns
        -------
        pynpoint.util.multiproc.TaskWriter
            Task writer.
        """

        return TaskWriter(self.m_result_queue,
                          image_out_port,
                          self.m_data_mutex)

    @typechecked
    def run(self) -> None:
        """
        Run method that starts the :class:`~pynpoint.util.multiproc.TaskCreator`, the instances
        of :class:`~pynpoint.util.multiproc.TaskProcessor`, and the
        :class:`~pynpoint.util.multiproc.TaskWriter`. They will be shut down when all tasks have
        finished.

        Returns
        -------
        NoneType
            None
        """

        # start all processes
        self.m_creator.start()

        for processor in self.m_task_processors:
            processor.start()

        self.m_writer.start()

        # wait for all tasks to have finished
        self.m_tasks_queue.join()
        self.m_result_queue.join()

        # clean up the processes
        for processor in self.m_task_processors:
            processor.join()

        self.m_writer.join()
        self.m_creator.join()


@typechecked
def apply_function(tmp_data: np.ndarray,
                   data_index: int,
                   func: Callable,
                   func_args: Optional[tuple]) -> np.ndarray:
    """
    Apply a function with optional arguments to the input data.

    Parameters
    ----------
    tmp_data : np.ndarray
        Input data.
    data_index : int
        Index of the data subset. When processing a stack of images, the argument of ``data_index``
        is the image index in the full stack.
    func : function
        Function.
    func_args : tuple, None
        Function arguments.

    Returns
    -------
    np.ndarray
        The results of the function.
    """

    if func_args is None:
        result = np.array(func(tmp_data, data_index))
    else:
        result = np.array(func(tmp_data, data_index, *func_args))

    return result


@typechecked
def to_slice(tuple_slice: tuple) -> tuple:
    """
    Function to convert tuples into slices for a multiprocessing queue.

    Parameters
    ----------
    tuple_slice : tuple
        Tuple to be converted into a slice.

    Returns
    -------
    tuple(slice, slice, slice)
        Tuple with three slices.
    """

    slices = []
    for item in tuple_slice:
        # slice(start, stop step)
        slices.append(slice(item[0], item[1], item[2]))

    return tuple(slices)



import os

import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.frameselection import RemoveFramesModule, FrameSelectionModule, \
                                               RemoveLastFrameModule, RemoveStartFramesModule, \
                                               ImageStatisticsModule, FrameSimilarityModule, \
                                               SelectByAttributeModule, ResidualSelectionModule
from pynpoint.util.tests import create_config, remove_test_data, create_star_data


class TestFrameSelection:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        create_star_data(self.test_dir+'images')
        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=['images'])

    def test_read_data(self) -> None:

        module = FitsReadingModule(name_in='read',
                                   image_tag='read',
                                   input_dir=self.test_dir+'images',
                                   overwrite=True,
                                   check=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('read')

        data = self.pipeline.get_data('read')
        assert np.sum(data) == pytest.approx(105.54278879805277, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

        attr = self.pipeline.get_attribute('read', 'NDIT', static=False)
        assert np.sum(attr) == pytest.approx(10, rel=self.limit, abs=0.)
        assert attr.shape == (2, )

        attr = self.pipeline.get_attribute('read', 'NFRAMES', static=False)
        assert np.sum(attr) == pytest.approx(10, rel=self.limit, abs=0.)
        assert attr.shape == (2, )

        self.pipeline.set_attribute('read', 'NDIT', [4, 4], static=False)

    def test_remove_last_frame(self) -> None:

        module = RemoveLastFrameModule(name_in='last',
                                       image_in_tag='read',
                                       image_out_tag='last')

        self.pipeline.add_module(module)
        self.pipeline.run_module('last')

        data = self.pipeline.get_data('last')
        assert np.sum(data) == pytest.approx(84.68885503527224, rel=self.limit, abs=0.)
        assert data.shape == (8, 11, 11)

        self.pipeline.set_attribute('last', 'PARANG', np.arange(8.), static=False)
        self.pipeline.set_attribute('last', 'STAR_POSITION', np.full((8, 2), 5.), static=False)

        attr = self.pipeline.get_attribute('last', 'PARANG', static=False)
        assert np.sum(attr) == pytest.approx(28., rel=self.limit, abs=0.)
        assert attr.shape == (8, )

        attr = self.pipeline.get_attribute('last', 'STAR_POSITION', static=False)
        assert np.sum(attr) == pytest.approx(80., rel=self.limit, abs=0.)
        assert attr.shape == (8, 2)

    def test_remove_start_frame(self) -> None:

        module = RemoveStartFramesModule(frames=1,
                                         name_in='start',
                                         image_in_tag='last',
                                         image_out_tag='start')

        self.pipeline.add_module(module)
        self.pipeline.run_module('start')

        data = self.pipeline.get_data('start')
        assert np.sum(data) == pytest.approx(64.44307047549808, rel=self.limit, abs=0.)
        assert data.shape == (6, 11, 11)

        attr = self.pipeline.get_attribute('start', 'PARANG', static=False)
        assert np.sum(attr) == pytest.approx(24., rel=self.limit, abs=0.)
        assert attr.shape == (6, )

        attr = self.pipeline.get_attribute('start', 'STAR_POSITION', static=False)
        assert np.sum(attr) == pytest.approx(60., rel=self.limit, abs=0.)
        assert attr.shape == (6, 2)

    def test_remove_frames(self) -> None:

        module = RemoveFramesModule(name_in='remove',
                                    image_in_tag='start',
                                    selected_out_tag='selected',
                                    removed_out_tag='removed',
                                    frames=[2, 5])

        self.pipeline.add_module(module)
        self.pipeline.run_module('remove')

        data = self.pipeline.get_data('selected')
        assert np.sum(data) == pytest.approx(43.68337741822863, rel=self.limit, abs=0.)
        assert data.shape == (4, 11, 11)

        data = self.pipeline.get_data('removed')
        assert np.sum(data) == pytest.approx(20.759693057269445, rel=self.limit, abs=0.)
        assert data.shape == (2, 11, 11)

        attr = self.pipeline.get_attribute('selected', 'PARANG', static=False)
        assert np.sum(attr) == pytest.approx(14., rel=self.limit, abs=0.)
        assert attr.shape == (4, )

        attr = self.pipeline.get_attribute('selected', 'STAR_POSITION', static=False)
        assert np.sum(attr) == pytest.approx(40., rel=self.limit, abs=0.)
        assert attr.shape == (4, 2)

        attr = self.pipeline.get_attribute('removed', 'PARANG', static=False)
        assert np.sum(attr) == pytest.approx(10., rel=self.limit, abs=0.)
        assert attr.shape == (2, )

        attr = self.pipeline.get_attribute('removed', 'STAR_POSITION', static=False)
        assert np.sum(attr) == pytest.approx(20., rel=self.limit, abs=0.)
        assert attr.shape == (2, 2)

    def test_frame_selection(self) -> None:

        module = FrameSelectionModule(name_in='select1',
                                      image_in_tag='start',
                                      selected_out_tag='selected1',
                                      removed_out_tag='removed1',
                                      index_out_tag='index1',
                                      method='median',
                                      threshold=2.,
                                      fwhm=0.1,
                                      aperture=('circular', 0.1),
                                      position=(None, None, 0.2))

        self.pipeline.add_module(module)
        self.pipeline.run_module('select1')

        data = self.pipeline.get_data('selected1')
        assert np.sum(data) == pytest.approx(54.58514780071149, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

        data = self.pipeline.get_data('removed1')
        assert np.sum(data) == pytest.approx(9.857922674786586, rel=self.limit, abs=0.)
        assert data.shape == (1, 11, 11)

        data = self.pipeline.get_data('index1')
        assert np.sum(data) == pytest.approx(5, rel=self.limit, abs=0.)
        assert data.shape == (1, )

        attr = self.pipeline.get_attribute('selected1', 'PARANG', static=False)
        assert np.sum(attr) == pytest.approx(17., rel=self.limit, abs=0.)
        assert attr.shape == (5, )

        attr = self.pipeline.get_attribute('selected1', 'STAR_POSITION', static=False)
        assert np.sum(attr) == pytest.approx(50, rel=self.limit, abs=0.)
        assert attr.shape == (5, 2)

        attr = self.pipeline.get_attribute('removed1', 'PARANG', static=False)
        assert np.sum(attr) == pytest.approx(7., rel=self.limit, abs=0.)
        assert attr.shape == (1, )

        attr = self.pipeline.get_attribute('removed1', 'STAR_POSITION', static=False)
        assert np.sum(attr) == pytest.approx(10, rel=self.limit, abs=0.)
        assert attr.shape == (1, 2)

        module = FrameSelectionModule(name_in='select2',
                                      image_in_tag='start',
                                      selected_out_tag='selected2',
                                      removed_out_tag='removed2',
                                      index_out_tag='index2',
                                      method='max',
                                      threshold=1.,
                                      fwhm=0.1,
                                      aperture=('annulus', 0.05, 0.1),
                                      position=None)

        self.pipeline.add_module(module)
        self.pipeline.run_module('select2')

        data = self.pipeline.get_data('selected2')
        assert np.sum(data) == pytest.approx(21.42652724866543, rel=self.limit, abs=0.)
        assert data.shape == (2, 11, 11)

        data = self.pipeline.get_data('removed2')
        assert np.sum(data) == pytest.approx(43.016543226832646, rel=self.limit, abs=0.)
        assert data.shape == (4, 11, 11)

        data = self.pipeline.get_data('index2')
        assert np.sum(data) == pytest.approx(10, rel=self.limit, abs=0.)
        assert data.shape == (4, )

        attr = self.pipeline.get_attribute('selected2', 'PARANG', static=False)
        assert np.sum(attr) == pytest.approx(8., rel=self.limit, abs=0.)
        assert attr.shape == (2, )

        attr = self.pipeline.get_attribute('selected2', 'STAR_POSITION', static=False)
        assert np.sum(attr) == pytest.approx(20, rel=self.limit, abs=0.)
        assert attr.shape == (2, 2)

        attr = self.pipeline.get_attribute('removed2', 'PARANG', static=False)
        assert np.sum(attr) == pytest.approx(16., rel=self.limit, abs=0.)
        assert attr.shape == (4, )

        attr = self.pipeline.get_attribute('removed2', 'STAR_POSITION', static=False)
        assert np.sum(attr) == pytest.approx(40, rel=self.limit, abs=0.)
        assert attr.shape == (4, 2)

        module = FrameSelectionModule(name_in='select3',
                                      image_in_tag='start',
                                      selected_out_tag='selected3',
                                      removed_out_tag='removed3',
                                      index_out_tag='index3',
                                      method='range',
                                      threshold=(10., 10.7),
                                      fwhm=0.1,
                                      aperture=('circular', 0.1),
                                      position=None)

        self.pipeline.add_module(module)
        self.pipeline.run_module('select3')

        data = self.pipeline.get_data('selected3')
        assert np.sum(data) == pytest.approx(22.2568501695632, rel=self.limit, abs=0.)
        assert data.shape == (2, 11, 11)

        data = self.pipeline.get_data('removed3')
        assert np.sum(data) == pytest.approx(42.18622030593487, rel=self.limit, abs=0.)
        assert data.shape == (4, 11, 11)

        data = self.pipeline.get_data('index3')
        assert np.sum(data) == pytest.approx(12, rel=self.limit, abs=0.)
        assert data.shape == (4, )

    def test_image_statistics_full(self) -> None:

        module = ImageStatisticsModule(name_in='stat1',
                                       image_in_tag='read',
                                       stat_out_tag='stat1',
                                       position=None)

        self.pipeline.add_module(module)
        self.pipeline.run_module('stat1')

        data = self.pipeline.get_data('stat1')
        assert np.sum(data) == pytest.approx(115.68591492205017, rel=self.limit, abs=0.)
        assert data.shape == (10, 6)

    def test_image_statistics_position(self) -> None:

        module = ImageStatisticsModule(name_in='stat2',
                                       image_in_tag='read',
                                       stat_out_tag='stat2',
                                       position=(5, 5, 0.1))

        self.pipeline.add_module(module)
        self.pipeline.run_module('stat2')

        data = self.pipeline.get_data('stat2')
        assert np.sum(data) == pytest.approx(118.7138708968444, rel=self.limit, abs=0.)
        assert data.shape == (10, 6)

    def test_frame_similarity_mse(self) -> None:

        module = FrameSimilarityModule(name_in='simi1',
                                       image_tag='read',
                                       method='MSE',
                                       mask_radius=(0., 0.1))

        self.pipeline.add_module(module)
        self.pipeline.run_module('simi1')

        attr = self.pipeline.get_attribute('read', 'MSE', static=False)
        assert np.min(attr) > 0.
        assert np.sum(attr) == pytest.approx(0.11739141370277852, rel=self.limit, abs=0.)
        assert attr.shape == (10, )

    def test_frame_similarity_pcc(self) -> None:

        module = FrameSimilarityModule(name_in='simi2',
                                       image_tag='read',
                                       method='PCC',
                                       mask_radius=(0., 0.1))

        self.pipeline.add_module(module)
        self.pipeline.run_module('simi2')

        attr = self.pipeline.get_attribute('read', 'PCC', static=False)
        assert np.min(attr) > 0.
        assert np.sum(attr) == pytest.approx(9.134820985662829, rel=self.limit, abs=0.)
        assert attr.shape == (10, )

    def test_frame_similarity_ssim(self) -> None:

        module = FrameSimilarityModule(name_in='simi3',
                                       image_tag='read',
                                       method='SSIM',
                                       mask_radius=(0., 0.1),
                                       temporal_median='constant')

        self.pipeline.add_module(module)
        self.pipeline.run_module('simi3')

        attr = self.pipeline.get_attribute('read', 'SSIM', static=False)
        assert np.min(attr) > 0.
        assert np.sum(attr) == pytest.approx(9.074290801266256, rel=self.limit, abs=0.)
        assert attr.shape == (10, )

    def test_select_by_attribute(self) -> None:

        self.pipeline.set_attribute('read', 'INDEX', np.arange(44), static=False)

        module = SelectByAttributeModule(name_in='frame_removal_1',
                                         image_in_tag='read',
                                         attribute_tag='SSIM',
                                         number_frames=6,
                                         order='descending',
                                         selected_out_tag='select_sim',
                                         removed_out_tag='remove_sim')

        self.pipeline.add_module(module)
        self.pipeline.run_module('frame_removal_1')

        attr = self.pipeline.get_attribute('select_sim', 'INDEX', static=False)
        assert np.sum(attr) == pytest.approx(946, rel=self.limit, abs=0.)
        assert attr.shape == (44, )

        attr = self.pipeline.get_attribute('select_sim', 'SSIM', static=False)
        assert np.sum(attr) == pytest.approx(5.545578246610884, rel=self.limit, abs=0.)
        assert attr.shape == (6, )

        attr = self.pipeline.get_attribute('remove_sim', 'SSIM', static=False)
        assert np.sum(attr) == pytest.approx(3.528712554655373, rel=self.limit, abs=0.)
        assert attr.shape == (4, )

    def test_residual_selection(self) -> None:

        module = ResidualSelectionModule(name_in='residual_select',
                                         image_in_tag='start',
                                         selected_out_tag='res_selected',
                                         removed_out_tag='res_removed',
                                         percentage=80.,
                                         annulus_radii=(0.1, 0.2))

        self.pipeline.add_module(module)
        self.pipeline.run_module('residual_select')

        data = self.pipeline.get_data('res_selected')
        assert np.sum(data) == pytest.approx(41.77295229983322, rel=self.limit, abs=0.)
        assert data.shape == (4, 11, 11)

        data = self.pipeline.get_data('res_removed')
        assert np.sum(data) == pytest.approx(22.670118175664847, rel=self.limit, abs=0.)
        assert data.shape == (2, 11, 11)



"""
Functions for adding attributes to a dataset in the central database.
"""

import warnings

import numpy as np

from astropy.io import fits
from typeguard import typechecked

from pynpoint.core.attributes import get_attributes
from pynpoint.core.dataio import ConfigPort, OutputPort


@typechecked
def set_static_attr(fits_file: str,
                    header: fits.header.Header,
                    config_port: ConfigPort,
                    image_out_port: OutputPort,
                    check: bool = True) -> None:
    """
    Function which adds the static attributes to the central database.

    Parameters
    ----------
    fits_file : str
        Name of the FITS file.
    header : astropy.io.fits.header.Header
        Header information from the FITS file that is read.
    config_port : pynpoint.core.dataio.ConfigPort
        Configuration port.
    image_out_port : pynpoint.core.dataio.OutputPort
        Output port of the images to which the static attributes are stored.
    check : bool
        Print a warning if certain attributes from the configuration file are not present in
        the FITS header. If set to `False`, attributes are still written to the dataset but
        there will be no warning if a keyword is not found in the FITS header.

    Returns
    -------
    NoneType
        None
    """

    attributes = get_attributes()

    static = []
    for key, value in attributes.items():
        if value['config'] == 'header' and value['attribute'] == 'static':
            static.append(key)

    for attr in static:

        fitskey = config_port.get_attribute(attr)

        if isinstance(fitskey, np.bytes_):
            fitskey = str(fitskey.decode('utf-8'))

        if fitskey != 'None':
            if fitskey in header:
                status = image_out_port.check_static_attribute(attr, header[fitskey])

                if status == 1:
                    image_out_port.add_attribute(attr, header[fitskey], static=True)

                elif status == 0:
                    pass

                elif status == -1:
                    warnings.warn(f'Static attribute {fitskey} has changed. Possibly the '
                                  f'current file {fits_file} does not belong to the data set '
                                  f'\'{image_out_port.tag}\'. Attribute value is updated.')

            elif check:
                warnings.warn(f'Static attribute {attr} (={fitskey}) not found in the FITS '
                              'header.')


@typechecked
def set_nonstatic_attr(header: fits.header.Header,
                       config_port: ConfigPort,
                       image_out_port: OutputPort,
                       check: bool = True) -> None:
    """
    Function which adds the non-static attributes to the central database.

    Parameters
    ----------
    header : astropy.io.fits.header.Header
        Header information from the FITS file that is read.
    config_port : pynpoint.core.dataio.ConfigPort
        Configuration port.
    image_out_port : pynpoint.core.dataio.OutputPort
        Output port of the images to which the non-static attributes are stored.

    Returns
    -------
    NoneType
        None
    """

    attributes = get_attributes()

    nonstatic = []
    for key, value in attributes.items():
        if value['attribute'] == 'non-static':
            nonstatic.append(key)

    for attr in nonstatic:
        if attributes[attr]['config'] == 'header':
            fitskey = config_port.get_attribute(attr)

            # if type(fitskey) == np.bytes_:
            #     fitskey = str(fitskey.decode('utf-8'))

            if fitskey != 'None':
                if fitskey in header:
                    image_out_port.append_attribute_data(attr, header[fitskey])

                elif header['NAXIS'] == 2 and attr == 'NFRAMES':
                    image_out_port.append_attribute_data(attr, 1)

                elif check:
                    warnings.warn('Non-static attribute %s (=%s) not found in the '
                                  'FITS header.' % (attr, fitskey))

                    image_out_port.append_attribute_data(attr, -1)


@typechecked
def set_extra_attr(fits_file: str,
                   nimages: int,
                   config_port: ConfigPort,
                   image_out_port: OutputPort,
                   first_index: int) -> int:
    """
    Function which adds extra attributes to the central database.

    Parameters
    ----------
    fits_file : str
        Absolute path and filename of the FITS file.
    nimages : int
        Number of images.
    config_port : pynpoint.core.dataio.ConfigPort
        Configuration port.
    image_out_port : pynpoint.core.dataio.OutputPort
        Output port of the images to which the attributes are stored.
    first_index : int
        First image index of the current subset.

    Returns
    -------
    int
        First image index for the next subset.
    """

    pixscale = config_port.get_attribute('PIXSCALE')

    image_index = np.arange(first_index, first_index+nimages, 1)

    for item in image_index:
        image_out_port.append_attribute_data('INDEX', item)

    image_out_port.append_attribute_data('FILES', fits_file)
    image_out_port.add_attribute('PIXSCALE', pixscale, static=True)

    return first_index + nimages
